"""
message.py  ─  言語別メッセージモジュールへのディスパッチャー

優先順位:
  1. set_language() で明示的に指定された言語
  2. システムの LOCALE（LC_ALL / LANG など）
  3. フォールバック: ' '.join(messages)
"""

import locale
import importlib
from typing import Union, Callable


def _fallback(messages: Union[tuple, str]) -> str:
    """言語モジュールが見つからない場合のデフォルト処理"""
    if isinstance(messages, tuple):
        return ' '.join(str(m) for m in messages)
    return str(messages)


_current_lang: str = ""
_to_message_func: Callable = _fallback


def _load_lang(lang: str) -> bool:
    """message_{lang}.py を動的にロードする。成功すれば True を返す"""
    global _to_message_func, _current_lang
    try:
        mod = importlib.import_module(f'.message_{lang}', package='yuichan')
        _to_message_func = mod.to_message
        _current_lang = lang
        return True
    except (ImportError, AttributeError):
        return False


def set_language(lang: str) -> bool:
    """
    メッセージ言語を切り替える。

    Parameters
    ----------
    lang : str
        言語コード（例: "ja", "en"）。
        対応するモジュール message_{lang}.py が存在しない場合は
        フォールバック（' '.join）のままになる。

    Returns
    -------
    bool
        モジュールのロードに成功すれば True
    """
    global _to_message_func, _current_lang
    ok = _load_lang(lang)
    if not ok:
        _to_message_func = _fallback
        _current_lang = ""
    return ok


def _detect_locale_lang() -> str:
    """システムの LOCALE から2文字の言語コードを返す。取得できなければ空文字"""
    try:
        # LC_ALL / LANG / LC_MESSAGES などを参照
        lang, _ = locale.getlocale()
        if lang:
            return lang[:2].lower()  # "ja_JP.UTF-8" → "ja"
    except Exception:
        pass
    return ""


def _init():
    """モジュールロード時に LOCALE から言語を自動検出して設定する"""
    lang = _detect_locale_lang()
    if lang:
        _load_lang(lang)   # 失敗してもフォールバックのまま続行


_init()


def to_message(messages: Union[tuple, str]) -> str:
    """
    YuiError の messages を現在の言語設定でメッセージ文字列に変換する。

    Parameters
    ----------
    messages : tuple | str
        YuiError.messages の値

    Returns
    -------
    str
        ユーザー向けのエラーメッセージ
    """
    return _to_message_func(messages)


def current_language() -> str:
    """現在ロードされている言語コードを返す。フォールバック中は空文字"""
    return _current_lang



"""
message_ja.py  ─  YuiError のエラーメッセージを初心者向けの日本語に変換するモジュール
"""

from typing import Union


# ──────────────────────────────────────────────────────────────
# ヘルパー
# ──────────────────────────────────────────────────────────────

def _extract(messages: tuple, prefix: str) -> list:
    """メッセージタプルから特定のプレフィクスを持つ項目をすべて抜き出す"""
    return [m[len(prefix):] for m in messages if isinstance(m, str) and m.startswith(prefix)]


def _bad(messages: tuple) -> str:
    """❌ の値（問題のある実際の値）。なければ空文字"""
    items = _extract(messages, "❌")
    return items[0] if items else ""


def _bads(messages: tuple) -> list:
    """❌ の値をすべてリストで返す"""
    return _extract(messages, "❌")


def _good(messages: tuple) -> str:
    """✅ の値（期待していた正しい値）。なければ空文字"""
    items = _extract(messages, "✅")
    return items[0] if items else ""


def _info(messages: tuple) -> str:
    """🔍 の値（補足情報）。なければ空文字"""
    items = _extract(messages, "🔍")
    return items[0] if items else ""


def _hint(messages: tuple) -> str:
    """⚠️ の値（ヒント）。なければ空文字"""
    items = _extract(messages, "⚠️")
    return items[0] if items else ""


def _key(messages: tuple, index: int) -> str:
    """タプルの index 番目の要素を返す。なければ空文字"""
    if isinstance(messages, tuple) and len(messages) > index:
        return messages[index]
    return ""


_TYPE_JA = {
    "number":   "数値",
    "int":      "整数",
    "float":    "小数",
    "string":   "文字列",
    "array":    "配列",
    "object":   "オブジェクト（辞書）",
    "data":     "データ",
    "function": "関数",
    "not-int":  "整数以外",
}


def _type_ja(raw: str) -> str:
    """✅<number> のような文字列から日本語型名を返す。例: '<number>' → '数値'"""
    name = raw.strip("<>")
    return _TYPE_JA.get(name, name)


# ──────────────────────────────────────────────────────────────
# メイン変換関数
# ──────────────────────────────────────────────────────────────

def to_message(messages: Union[tuple, str]) -> str:
    """
    YuiError の messages を初心者向けの日本語メッセージに変換する。

    Parameters
    ----------
    messages : tuple | str
        YuiError.messages の値

    Returns
    -------
    str
        親しみやすい日本語のエラーメッセージ
    """
    if isinstance(messages, str):
        if messages == 'interruptted':
            return "実行が中断されたよ。"
        return messages

    if not isinstance(messages, tuple) or len(messages) == 0:
        return f"なんかエラーが起きちゃったよ… ({messages})"

    k0 = _key(messages, 0)
    k1 = _key(messages, 1)
    bad  = _bad(messages)
    good = _good(messages)
    info = _info(messages)
    hint = _hint(messages)

    # ────────────────────────────────────────────────
    # "error" カテゴリ（ランタイム）
    # ────────────────────────────────────────────────

    if k0 == "error":

        if k1 == "type":
            type_ja = _type_ja(good) if good else "別の型"
            if bad:
                return (
                    f"{type_ja}が必要なところに「{bad}」が来ているよ。\n"
                    f"ここには{type_ja}を使ってみてね！"
                )
            return f"ここには{type_ja}が必要だよ！"

        if k1 == "index":
            limit = good.lstrip("<") if good else ""  # "✅<5" → "5"
            if bad and limit:
                return (
                    f"配列の {bad} 番目にアクセスしようとしたけど、存在しないよ！\n"
                    f"このリストは {limit} 個の要素しかないよ（番号は 0 から始まるよ）。"
                )
            if bad:
                return (
                    f"配列の {bad} 番目にアクセスしようとしたけど、その番号は存在しないよ！\n"
                    "番号は 0 から始まるよ。配列の大きさを確認してみてね。"
                )
            return (
                "配列の範囲外にアクセスしようとしているよ！\n"
                "番号が大きすぎないか確認してみてね（番号は 0 から始まるよ）。"
            )

        if k1 == "incomparable":
            bads = _bads(messages)
            if len(bads) >= 2:
                return (
                    f"「{bads[0]}」と「{bads[1]}」は比べられないよ〜。\n"
                    "同じ種類のデータ（数字どうし・文字列どうし）じゃないと比べられないよ！"
                )
            return "比べられない値どうしを比較しようとしているよ。同じ種類のデータか確認してね！"

        if k1 == "recursion":
            if info:
                return (
                    f"関数「{info}」が自分自身を呼び出しすぎてしまったよ（再帰が深すぎる）！\n"
                    "再帰を止める条件（ベースケース）を書き忘れていないかな？"
                )
            return (
                "関数が自分自身を呼び出しすぎてしまったよ（再帰が深すぎる）！\n"
                "再帰を止める条件（ベースケース）を書き忘れていないかな？"
            )

        if k1 == "timeout":
            if bad:
                return (
                    f"実行が {bad} 秒を超えて時間切れになっちゃったよ〜⏰\n"
                    "無限ループになっていないか確認してみてね！\n"
                    "ループを抜ける条件が正しく書けているかな？"
                )
            return (
                "実行が時間切れになっちゃったよ〜⏰\n"
                "無限ループになっていないか確認してみてね！"
            )

        if k1 == "division by zero":
            if bad:
                return (
                    f"0 で割り算しようとしているよ！{bad} を使っているところを確認してね。\n"
                    "0 で割ることはできないよ。"
                )
            return "0 で割り算しようとしているよ！0 で割ることはできないよ。"

        if k1 == "conversion":
            if bad:
                return (
                    f"「{bad}」を数値に変換できなかったよ。\n"
                    "数字として読めない文字が含まれていないか確認してみてね！"
                )
            return "値を数値に変換できなかったよ。数字として読める文字列か確認してね！"

        if k1 == "internal":
            detail = info or hint
            if detail:
                return (
                    f"内部エラーが発生しちゃったよ（バグかも）。\n"
                    f"詳細: {detail}"
                )
            return "内部エラーが発生しちゃったよ（バグかも）。開発者に教えてあげてね！"

        # error + unknown
        return f"エラーが起きちゃったよ（{k1}）: " + " ".join(
            m for m in messages if isinstance(m, str)
        )

    # ────────────────────────────────────────────────
    # "undefined" カテゴリ
    # ────────────────────────────────────────────────

    if k0 == "undefined":

        if k1 == "variable":
            if bad:
                return (
                    f"変数「{bad}」はまだ定義されていないよ！\n"
                    f"使う前に「{bad} = ...」のように値を入れてあげてね。"
                )
            return "まだ定義されていない変数を使おうとしているよ。先に値を入れてね！"

        if k1 == "function":
            if bad:
                return (
                    f"関数「{bad}」が見つからないよ〜。\n"
                    "名前のスペルが違うか、定義する前に呼び出しているかも！"
                )
            return "呼び出した関数が見つからないよ。名前を確認してみてね！"

    # ────────────────────────────────────────────────
    # "expected" カテゴリ（パーサー・ランタイム）
    # ────────────────────────────────────────────────

    if k0 == "expected":

        if k1 == "token":
            if good and bad:
                return (
                    f"ここには「{good}」が必要なのに「{bad}」って書いてあったよ。\n"
                    "書き方を確認してみてね！"
                )
            if good:
                return f"ここには「{good}」が必要だよ！"
            return "必要な記号や文字が足りないよ。書き方を確認してみてね！"

        if k1 == "closing":
            if good:
                return (
                    f"開いたカッコや括弧を閉じ忘れているよ！「{good}」で閉じてね。\n"
                    "対応する括弧がちゃんとペアになっているか確認してみて！"
                )
            return "開いたカッコや括弧を閉じ忘れているよ！対応する括弧を確認してね。"

        if k1 == "variable":
            if bad:
                return (
                    f"「{bad}」に代入しようとしたけど、これは変数じゃないみたい。\n"
                    "代入できるのは変数名だけだよ！"
                )
            return "変数以外のものに代入しようとしているよ。代入先を確認してみてね！"

        if k1 == "arguments":
            if good and bad:
                return (
                    f"関数の引数の数が違うよ！\n"
                    f"必要な数は {good} 個なのに、{bad} 個渡しているよ。\n"
                    "引数の数を合わせてみてね！"
                )
            return "関数の引数の数が違うよ。引数の数を確認してみてね！"

        if k1 == "boolean":
            return (
                "true（はい）か false（いいえ）が必要なところだよ。\n"
                "条件式の書き方を確認してみてね！"
            )

        if k1 == "number":
            if bad:
                return (
                    f"数字が必要なところなのに「{bad}」って書いてあったよ。\n"
                    "数字を正しく書いてみてね！"
                )
            return "数字が必要なところだよ。数字を書いてみてね！"

        if k1 == "string":
            if bad:
                return (
                    f"文字列が必要なところなのに「{bad}」って書いてあったよ。\n"
                    '文字列はダブルクォート（""）で囲んでね！'
                )
            return '文字列が必要なところだよ。ダブルクォート（""）で囲んでね！'

        if k1 == "array":
            return (
                "配列が必要なところだよ。\n"
                "[1, 2, 3] のように [ ] で囲んで書いてみてね！"
            )

        if k1 == "object":
            return (
                "オブジェクト（辞書）が必要なところだよ。\n"
                '{"key": value} のように { } で囲んで書いてみてね！'
            )

        # 汎用ノンターミナル: ("expected", "funcdef", "❌...", "⚠️...")
        if bad:
            label = k1 if k1 else "構文"
            msg = f"「{label}」の書き方が間違っているみたい。「{bad}」って書いてあったよ。"
            if hint:
                msg += f"\nヒント: {hint}"
            return msg

        return f"ここに「{k1}」が必要だよ。書き方を確認してみてね！"

    # ────────────────────────────────────────────────
    # "required" カテゴリ
    # ────────────────────────────────────────────────

    if k0 == "required" and k1 == "arguments":
        if bad:
            return (
                f"この関数には引数が必要なのに、{bad} 個しか渡していないよ。\n"
                "引数を渡して呼び出してみてね！"
            )
        return "この関数には引数が必要だよ。引数を渡して呼び出してみてね！"

    # ────────────────────────────────────────────────
    # "wrong" カテゴリ（パーサー）
    # ────────────────────────────────────────────────

    if k0 == "wrong":

        if k1 == "statement":
            if bad:
                return (
                    f"「{bad}」はゆいの文として知らないよ〜。\n"
                    "スペルや書き方が間違っているかもしれないよ！"
                )
            return "この書き方はゆいの文として知らないよ〜。書き方を確認してみてね！"

        if k1 == "token":
            if bad and good:
                return (
                    f"ここは「{good}」のはずなのに「{bad}」って書いてあったよ。\n"
                    "書き方を確認してみてね！"
                )
            return "ここには違うものが書いてあるよ。書き方を確認してみてね！"

        if k1 == "name":
            if bad:
                return (
                    f"「{bad}」は変数名や関数名として使えないよ。\n"
                    "名前はアルファベット・ひらがな・漢字などで始めて、\n"
                    "数字や記号は使えないよ！"
                )
            return "変数名・関数名の書き方がおかしいよ。名前の付け方を確認してみてね！"

        if k1 == "indent":
            if good:
                return (
                    f"インデント（字下げ）がおかしいよ。\n"
                    f"ブロックの終わりは「{good}」のインデントに合わせてね！"
                )
            return (
                "インデント（字下げ）がおかしいよ。\n"
                "ブロックの中はそろえて字下げしてね！"
            )

    # ────────────────────────────────────────────────
    # "bad" カテゴリ（パーサー）
    # ────────────────────────────────────────────────

    if k0 == "bad":

        if k1 == "escape sequence":
            return (
                "バックスラッシュ（\\）の後に変な文字があるよ。\n"
                "使えるのは \\n（改行）・\\t（タブ）・\\\\（\\自身）などだよ！"
            )

    # ────────────────────────────────────────────────
    # "mismatch" カテゴリ
    # ────────────────────────────────────────────────

    if k0 == "mismatch" and k1 == "arguments":
        if good and bad:
            return (
                f"関数の引数の数が違うよ！\n"
                f"必要な数は {good} 個なのに、{bad} 個渡しているよ。\n"
                "引数の数を合わせてみてね！"
            )
        return "関数の引数の数が違うよ。引数の数を確認してみてね！"

    # ────────────────────────────────────────────────
    # "failed" カテゴリ（テスト失敗）
    # ────────────────────────────────────────────────

    if k0 == "failed" and k1 == "test":
        test_expr = info  # 🔍 に式が入っている
        if bad and good:
            expr_line = f"「{test_expr}」の結果を確認してみてね。\n" if test_expr else ""
            return (
                f"テスト失敗！{expr_line}"
                f"  期待した値: {good}\n"
                f"  実際の値:   {bad}"
            )
        return f"テストが失敗したよ！" + (f"「{test_expr}」の結果を確認してみてね。" if test_expr else "")

    # ────────────────────────────────────────────────
    # 文字列エラー（interruptted など）
    # ────────────────────────────────────────────────

    if k0 == "interruptted":
        return "実行が中断されたよ。"

    # ────────────────────────────────────────────────
    # フォールバック
    # ────────────────────────────────────────────────
    parts = [m for m in messages if isinstance(m, str)]
    return "エラーが起きちゃったよ: " + " ".join(parts)



from dataclasses import dataclass
from typing import List, Optional, Dict, Any, Union
from abc import ABC, abstractmethod

from .yuiast import ASTNode, set_operators


class YuiError(RuntimeError):
    """Yui言語のエラーを表現するクラス"""
    messages: tuple
    error_node: Optional[ASTNode]
    runtime: Optional['YuiRuntime'] # type: ignore
    BK: bool
    
    def __init__(self, messages: tuple, error_node: Optional[ASTNode] = None,
                 runtime: Optional['YuiRuntime'] = None, #type: ignore
                 BK: bool = False):
        """YuiErrorを初期化する"""
        super().__init__(' '.join(messages))
        self.messages = messages
        self.error_node = error_node if isinstance(error_node, ASTNode) else None
        self.runtime = runtime
        self.BK = BK

    @property
    def lineno(self) -> int:
        """エラー箇所の行番号を返す（1始まり）"""
        if self.error_node:
            line, _, _ = self.error_node.extract()
            return line
        return 0

    @property
    def offset(self) -> int:
        """エラー箇所の列番号を返す（1始まり）"""
        if self.error_node:
            _, offset, _ = self.error_node.extract()
            return offset
        return 0

    @property
    def text(self) -> str:
        """エラー箇所のコードスニペットを返す"""
        if self.error_node:
            _, _, snippet = self.error_node.extract()
            return snippet
        return ""

    def formatted_message(self, prefix=" ", marker: str = '^', lineoffset: int = 0) -> str:
        """エラーメッセージを整形して返す"""
        # message = _to_message(self.messages)
        message = ' '.join(self.messages)   
        if self.error_node:
            line, col, snippet = self.error_node.extract()
            # エラー範囲の長さを計算（最小3文字）
            length = max(self.error_node.end_pos - self.error_node.pos, 3) if self.error_node.end_pos is not None else 3
            # エラー位置を指すポインタを作成
            make_pointer = marker * min(length, 16)
            snippet = snippet.split('\n')[0]  # エラー行の最初の行だけを表示
            indent = " " * (col - 1)
            message = f"{message} line {line + lineoffset}, column {col}:\n{prefix}{snippet}\n{prefix}{indent}{make_pointer}"
        if self.runtime is None:
            message = f"[構文エラー/SyntaxError] {message}"
        else:
            message = f"[実行時エラー/RuntimeError] {message}\n[環境/Environment] {self.runtime.stringfy_env(stack=-1)}\n"
        return message

TY_NULL = '⛔'
TY_BOOLEAN = '🔘'
TY_INT = '💯'
TY_FLOAT = '📊'
TY_NUMBER = '🔢'
TY_ARRAY = '🍡'
TY_OBJECT = '🗂️'
TY_STRING = '💬'

@dataclass
class YuiType(ABC):
    """Yui言語の型を表現するクラス"""
    name: str
    emoji: str

    def __init__(self, name: str, emoji: str):
        self.name = name
        self.emoji = emoji

    def __str__(self):
        return self.emoji

    @abstractmethod
    def match(self, value: Any) -> bool:
        pass

    def match_or_raise(self, node_or_value = None):
        value = node_or_value
        if not self.match(value):
            raise YuiError(("error", "type", f"✅<{self.emoji}{self.name}>", f"❌{value}"), node_or_value)

    @abstractmethod
    def to_arrayview(self, n: int) -> List[int]:
        pass

    @abstractmethod
    def to_native(self, elements: List[int], node=None) -> Any:
        pass

    @abstractmethod
    def stringfy(self, native_value: Any, indent_prefix: str = "", width=80) -> str:
        pass
    
    def equals(self, left_node: Any, right_node: Any) -> bool:
        left_value = YuiType.to_native((left_node))
        right_value = YuiType.to_native((right_node))
        return left_value == right_value    

    def less_than(self, left_node: Any, right_node: Any, op = "<") -> bool:
        left_value = left_node
        right_value = right_node
        raise YuiError(("unsupported", "comparison", f"❌{left_value} {op} {right_value}"), left_node)


    # クラス変数として各型のインスタンスを定義
    NullType: 'YuiNullType' = None
    BooleanType: 'YuiBooleanType' = None
    IntType: 'YuiIntType' = None
    FloatType: 'YuiFloatType' = None
    NumberType: 'YuiNumberType' = None
    StringType: 'YuiStringType' = None
    ArrayType: 'YuiArrayType' = None
    ObjectType: 'YuiObjectType' = None

    @staticmethod
    def evaluated(value: Any) -> Any:
        return value

    @staticmethod
    def is_bool(node_or_value: Any) -> bool:
        return YuiType.BooleanType.match(node_or_value)

    @staticmethod
    def is_int(node_or_value: Any) -> bool:
        return YuiType.IntType.match(node_or_value)

    @staticmethod
    def is_float(node_or_value: Any) -> bool:
        return YuiType.FloatType.match(node_or_value)

    @staticmethod
    def is_number(node_or_value: Any) -> bool:
        return YuiType.NumberType.match(node_or_value)      
    
    @staticmethod
    def is_string(node_or_value: Any) -> bool:
        return YuiType.StringType.match(node_or_value)

    @staticmethod
    def is_array(node_or_value: Any) -> bool:
        return YuiType.ArrayType.match(node_or_value)

    @staticmethod
    def is_object(node_or_value: Any) -> bool:
        return YuiType.ObjectType.match(node_or_value)

    @staticmethod
    def matched_native(node_or_value) -> None:
        value = node_or_value
        if isinstance(value, YuiValue):
            return value.native
        return value # 型チェック済みを想定

    @staticmethod
    def from_arrayview(value: Any) -> 'YuiValue':
        if isinstance(value, YuiValue):
            return value
        return YuiValue(value)

    @staticmethod
    def into_arrayview(node_or_value) -> Any:
        value = node_or_value
        if isinstance(value, YuiValue) and value.is_primitive():
            return value.native
        if isinstance(value, (int, float, str)) or value is None:
            return value
        return YuiValue(value)

    @staticmethod
    def arrayview_s(value):
        if value is None:
            return "null"
        elif isinstance(value, str):
            value = value.replace('"', '\\"').replace('\n', '\\n')
            return f'"{value}"'
        elif isinstance(value, float):
            return f"{value:.6f}"
        else:
            return str(value)
    
    @staticmethod
    def to_native(node_or_value) -> Any:
        value = node_or_value
        if isinstance(value, YuiValue):
            return value.native
        if isinstance(value, list):
            return [YuiType.to_native(v) for v in value]
        if isinstance(value, dict):
            return {str(k): YuiType.to_native(v) for k, v in value.items()}
        return value

    @staticmethod
    def native_to_yui(native_value) -> Any:
        """ネイティブ値をYui形式に変換して返す"""
        if native_value is None:
            return YuiValue.NullValue
        if isinstance(native_value, bool):
            return YuiValue.TrueValue if native_value else YuiValue.FalseValue
        if isinstance(native_value, (int, float, list, str, dict, tuple)):
            return YuiValue(native_value)
        assert isinstance(native_value, YuiValue)
        return native_value

    @staticmethod
    def yui_to_native(value: Any) -> Any:
        """Yui形式の値をネイティブ値に変換して返す"""
        if isinstance(value, YuiValue):
            return value.native
        if isinstance(value, list):
            return [YuiType.yui_to_native(e) for e in value]
        if isinstance(value, dict):
            new_value = {}
            for key, item in value.items():
                new_value[key] = YuiType.yui_to_native(item)
            return new_value
        return value

    @staticmethod
    def compare(left_node_or_value, right_node_or_value: Any) -> int:
        print(f"Comparing {left_node_or_value} and {right_node_or_value}")
        if YuiType.is_number(left_node_or_value) and YuiType.is_number(right_node_or_value):
            left_value = round(YuiType.matched_native(left_node_or_value), 6)
            right_value = round(YuiType.matched_native(right_node_or_value), 6)
            return _compare(left_value, right_value)
        if YuiType.is_string(left_node_or_value) and YuiType.is_string(right_node_or_value):
            left_value = YuiType.matched_native(left_node_or_value)
            right_value = YuiType.matched_native(right_node_or_value)
            return _compare(left_value, right_value)
        left_value = left_node_or_value
        right_value = right_node_or_value        
        if not isinstance(right_value, YuiValue):
            right_value = YuiValue(right_value.native)
        return _compare(left_value.arrayview, right_value.arrayview)

def _compare(left, right) -> int:
    if left == right:
        return 0
    if left < right:
        return -1
    else:
        return 1

class YuiNullType(YuiType):
    def __init__(self):
        super().__init__("null", TY_NULL)

    def match(self, node_or_value: Any) -> bool:
        value = node_or_value
        return value is None or (isinstance(value, YuiValue) and isinstance(value.type, YuiNullType))

    def check_element(self, node_or_value: Any) -> None:
        raise YuiError(("immutable", f"❌{self}"), node_or_value)

    def to_arrayview(self, n: None) -> List[int]:
        return []

    def to_native(self, elements: List[int], node=None) -> None:
        return None
    
    def stringfy(self, native_value: None, indent_prefix: str = "", width=80) -> str:
        return "null"

class YuiBooleanType(YuiType):
    def __init__(self):
        super().__init__("boolean", TY_BOOLEAN)

    def match(self, node_or_value: Any) -> bool:
        value = node_or_value
        # Python の bool は int のサブクラスなので先に isinstance(value, bool) でチェックする
        return isinstance(value, bool) or (isinstance(value, YuiValue) and isinstance(value.type, YuiBooleanType))

    def check_element(self, node_or_value: Any) -> None:
        raise YuiError(("immutable", f"❌{self}"), node_or_value)

    def to_arrayview(self, n: bool) -> List[int]:
        return [1] if n else []

    def to_native(self, elements: List[int], node=None) -> bool:
        return len(elements) > 0

    def stringfy(self, native_value: bool, indent_prefix: str = "", width=80) -> str:
        return "true" if native_value else "false"

    def equals(self, left_node: Any, right_node: Any) -> bool:
        left_value = YuiType.to_native(left_node)
        right_value = YuiType.to_native(right_node)
        if isinstance(right_value, bool):
            return left_value == right_value
        return False

    def less_than(self, left_node: Any, right_node: Any, op: str = "<") -> bool:
        left_value = YuiType.to_native(left_node)
        right_value = YuiType.to_native(right_node)
        if isinstance(right_value, bool):
            return left_value < right_value
        return super().less_than(left_node, right_node, op=op)


class YuiIntType(YuiType):
    def __init__(self):
        super().__init__("int", TY_INT)

    def match(self, node_or_value: Any) -> bool:
        value = node_or_value
        return isinstance(value, int) or (isinstance(value, YuiValue) and isinstance(value.type, YuiIntType))

    def check_element(self, node_or_value: Any) -> None:
        YuiType.IntType.match_or_raise(node_or_value)
        value = self.matched_native(node_or_value)
        if value != 0 and value != 1:
            raise YuiError(("error", "value", f"✅0/1", f"❌{value}"), node_or_value)

    def to_arrayview(self, n: int) -> List[int]:
        """整数を32ビットの2の補数表現の整数配列に変換"""
        # 32ビットにマスク（符号なし整数として扱う）
        n_unsigned = n & 0xFFFFFFFF    
        # 各ビットを抽出（MSBからLSBへ）
        bits = []
        for i in range(31, -1, -1):
            value = (n_unsigned >> i) & 1
            bits.append(value) # int はそのまま
        return bits

    def to_native(self, bits: List[int], node=None) -> int:
        """32ビットの2の補数表現の整数配列を整数に変換"""
        if len(bits) != 32:
            raise YuiError(("array", "format", f"❌{len(bits)}", "✅32"), node)
        
        # ビット列を符号なし整数に変換
        n_unsigned = 0
        for bit in bits:
            n_unsigned = (n_unsigned << 1) | bit
        
        # 32ビット符号付き整数に変換（2の補数）
        if n_unsigned >= 0x80000000:  # MSBが1（負数）
            n = n_unsigned - 0x100000000
        else:
            n = n_unsigned        
        return n
    
    def stringfy(self, native_value: int, indent_prefix: str = "", width=80) -> str:
        return f"{native_value}"
    
    def equals(self, left_node: Any, right_node: Any) -> bool:
        left_value = YuiType.to_native(left_node)
        right_value = YuiType.to_native(right_node)
        if isinstance(right_value, bool):
            return False  # bool は IntType と等しくない
        if isinstance(right_value, float):
            return round(float(left_value), 6) == round(right_value, 6)
        return left_value == right_value

    def less_than(self, left_node: Any, right_node: Any, op = "<") -> bool:
        left_value = YuiType.to_native(left_node)
        right_value = YuiType.to_native(right_node)
        if isinstance(right_value, float):
            return round(float(left_value), 6) < round(right_value, 6)
        if isinstance(right_value, int):
            return left_value < right_value
        super().less_than(left_node, right_node, op=op)
    

class YuiFloatType(YuiType):
    def __init__(self):
        super().__init__("float", TY_FLOAT)

    def match(self, node_or_value: Any) -> bool:
        value = node_or_value
        return isinstance(value, float) or (isinstance(value, YuiValue) and isinstance(value.type, YuiFloatType))

    def check_element(self, node_or_value: Any) -> None:
        YuiType.IntType.match_or_raise(node_or_value)
        value = self.matched_native(node_or_value)
        if value < -1 or value > 9:
            raise YuiError(("error", "value", f"✅-1,0-9", f"❌{value}"), node_or_value)

    def to_arrayview(self, x: float) -> List[int]:
        """浮動小数点数を符号付き一桁整数配列に変換
        Example:
            [1, 3, 1, 4, 0, 0, 0, 0]  # 3.140000
            [-1, 2, 5, 0, 0, 0, 0, 0]  # -2.500000
        """
        sign = -1 if x < 0 else 1
        s = f"{abs(x):.6f}"  # 小数点以下6桁に丸める
        s = s.replace('.', '')  # 小数点を削除
        digits = [sign] + [int(ch) for ch in s]
        return digits

    def to_native(self, digits: List[int], node=None) -> float:
        """符号付き一桁整数配列を浮動小数点数に変換
        Example:
            >>> array_to_float([1, 3, 1, 4, 0, 0, 0, 0])
            3.140000
            >>> array_to_float([-1, 2, 5, 0, 0, 0, 0, 0])
            -2.500000
        """
        sign = digits[0]
        if not(isinstance(sign, int) and sign in (1, -1)):
            raise YuiError(f"conversion", "tofloat", f"❌[0]{digits[0]}", f"✅1/-1", f"🔍{digits}", node)
        num_digits = digits[1:]
        for i, d in enumerate(num_digits, start=1):
            if not (isinstance(d, int) and 0 <= d <= 9):
                raise YuiError(f"conversion", "tofloat", f"❌[{i}]{d}", f"✅0-9", f"🔍{digits}", node)
        s = ''.join(str(d) for d in num_digits)
        if len(s) <= 6:
            # 小数部のみの場合（整数部なし）
            value = int(s)
        else:
            # 小数点を6桁前に挿入
            value = float(s[:-6] + '.' + s[-6:])
        return sign * value
    
    def stringfy(self, native_value: float, indent_prefix: str = "", width=80) -> str:
        return f"{native_value:.6f}"

    def equals(self, left_node: Any, right_node: Any) -> bool:
        left_value = round(YuiType.to_native(left_node), 6)
        right_value = YuiType.to_native((right_node))
        if isinstance(right_value, (float, int)):
            return left_value == round(right_value, 6)
        return False   

    def less_than(self, left_node: Any, right_node: Any, op = "<") -> bool:
        left_value = YuiType.to_native(left_node)
        right_value = YuiType.to_native(right_node)
        if isinstance(right_value, (int, float)):
            return left_value < right_value
        super().less_than(left_node, right_node, op=op)

class YuiNumberType(YuiType):
    def __init__(self):
        super().__init__("number", TY_NUMBER)

    def match(self, node_or_value: Any) -> bool:
        return YuiType.IntType.match(node_or_value) or YuiType.FloatType.match(node_or_value)

    def check_element(self, node_or_value: Any) -> None:
        pass

    def to_arrayview(self, n: Union[int, float]) -> List[int]:
        if isinstance(n, float):
            return YuiType.FloatType.to_arrayview(n)
        else:
            return YuiType.IntType.to_arrayview(n)
    
    def to_native(self, bits: List[int], node=None) -> int:
        if len(bits) == 32:
            return YuiType.IntType.to_native(bits, node=node)
        else:
            return YuiType.FloatType.to_native(bits, node=node)
    
    def stringfy(self, native_value: int, indent_prefix: str = "", width=80) -> str:
        if isinstance(native_value, float):
            return YuiType.FloatType.stringfy(native_value, indent_prefix=indent_prefix, width=width)
        else:
            return YuiType.IntType.stringfy(native_value, indent_prefix=indent_prefix, width=width)

class YuiStringType(YuiType):
    def __init__(self):
        super().__init__("string", TY_STRING)

    def match(self, node_or_value: Any) -> bool:
        value = node_or_value
        return isinstance(value, str) or (isinstance(value, YuiValue) and isinstance(value.type, YuiStringType))

    def check_element(self, node_or_value: Any) -> None:
        YuiType.IntType.match_or_raise(node_or_value)

    def to_arrayview(self, x: str) -> List[int]:
        """文字コード"""
        return [ord(ch) for ch in x]

    def to_native(self, elements: List[int], node=None) -> str:
        return ''.join(chr(d) for d in elements)

    def stringfy(self, native_value: str, indent_prefix: str = "", width=80) -> str:
        content = native_value.replace('"', '\\"').replace('\n', '\\n')
        return f'"{content}"'
    
    def equals(self, left_node: Any, right_node: Any) -> bool:
        left_value = YuiType.to_native(left_node)
        if YuiType.is_string(right_node):
            right_value = YuiType.matched_native(right_node)
            return left_value == right_value
        return False   

    def less_than(self, left_node: Any, right_node: Any, op = "<") -> bool:
        left_value = YuiType.to_native(left_node)
        if YuiType.is_string(right_node):
            right_value = YuiType.matched_native(right_node)
            return left_value < right_value
        return False

def _array_equal(a, b) -> bool:
    """配列の再帰的等価比較（文字コード配列と文字列の相互比較を含む）"""
    if isinstance(a, list) and isinstance(b, list):
        if len(a) != len(b):
            return False
        return all(_array_equal(x, y) for x, y in zip(a, b))
    if isinstance(a, list) and isinstance(b, str):
        try:
            return ''.join(chr(c) for c in a) == b
        except (TypeError, ValueError):
            return False
    if isinstance(a, str) and isinstance(b, list):
        return _array_equal(b, a)
    return a == b


class YuiArrayType(YuiType):
    def __init__(self):
        super().__init__("array", TY_ARRAY)

    def match(self, node_or_value: Any) -> bool:
        value = node_or_value
        return isinstance(value, list) or (isinstance(value, YuiValue) and isinstance(value.type, YuiArrayType))
    
    def check_element(self, node_or_value: Any) -> None:
        pass

    def to_arrayview(self, array_value: list) -> List[int]:
        """配列の要素をエンコード"""
        return [YuiType.into_arrayview(value) for value in array_value]

    def to_native(self, elements: List[int], node=None) -> str:
        array = []
        for element in elements:
            if isinstance(element, YuiValue):
                array.append(element.native)
            else:
                array.append(element)
        return array
    
    def stringfy(self, elements: List[int], indent_prefix: str = "", width=80) -> str:
        buffer = ["["]
        for i, element in enumerate(elements):
            if i > 0:
                buffer.append(", ")
            buffer.append(f"{YuiType.arrayview_s(element)}")
        buffer.append("]")
        string_content = ''.join(buffer)
        if indent_prefix is None or len(indent_prefix) + len(string_content) <= width:
            return string_content
        inner_indent_prefix = indent_prefix + "  "
        LF = "\n"
        buffer = ["["]
        for i, element in enumerate(elements):
            buffer.append(f"{LF}{inner_indent_prefix}")
            if isinstance(element, YuiValue):
                buffer.append(element.stringfy(inner_indent_prefix, width=width))
            else:
                buffer.append(f"{YuiType.arrayview_s(element)}")
            if i < len(elements) - 1:
                buffer.append(",")
        buffer.append(f"{LF}{indent_prefix}]")
        return ''.join(buffer)

    def equals(self, left_node: Any, right_node: Any) -> bool:
        left_native = YuiType.to_native(left_node)
        right_native = YuiType.to_native(right_node)
        return _array_equal(left_native, right_native)


class YuiObjectType(YuiType):
    def __init__(self):
        super().__init__("object", TY_OBJECT)

    def match(self, node_or_value: Any) -> bool:
        value = node_or_value
        return isinstance(value, dict) or (isinstance(value, YuiValue) and isinstance(value.type, YuiObjectType))

    def check_element(self, node_or_value: Any) -> None:
        global ArrayType, StringType
        ArrayType.match_or_raise(node_or_value)
        array = self.matched_native(node_or_value)
        if len(array) != 2 or not StringType.match(array[0]):
            raise YuiError(("error", "value", f"✅[key, value]", f"❌{array}"), node_or_value)

    def to_arrayview(self, object_value: dict) -> List[int]:
        """オブジェクトのキーと値をエンコード"""
        elements = []
        for key, value in object_value.items():
            elements.append(YuiValue([str(key), YuiType.into_arrayview(value)]))
        return elements

    def to_native(self, elements: List[int], node=None) -> str:
        obj = {}
        for key_value in elements:
            if not isinstance(key_value, YuiValue):
                raise YuiError(f"conversion", "toobject", f"❌{key_value}", f"✅[key, value]", f"🔍{elements}", node)
            key_value = key_value.native
            if not isinstance(key_value, list) or len(key_value) != 2:
                raise YuiError(f"conversion", "toobject", f"❌{key_value}", f"✅[key, value]", f"🔍{elements}", node)
            key = key_value[0]
            if not isinstance(key, str):
                raise YuiError(f"conversion", "toobject", f"❌{key}", f"✅<string>", f"🔍{key_value}", node)
            value = key_value[1]
            obj[key] = value
        return obj

    def stringfy(self, native_value: dict, indent_prefix: str = "", width=80) -> str:
        buffer = ["{"]
        for i, (key, value) in enumerate(native_value.items()):
            if i > 0:
                buffer.append(", ")
            buffer.append(f'"{key}": {YuiType.arrayview_s(value)}')
        buffer.append("}")
        string_content = ''.join(buffer)
        if indent_prefix is None or len(indent_prefix) + len(string_content) <= width:
            return string_content
        inner_indent_prefix = indent_prefix + "  "
        LF = "\n"
        buffer = ["{"]
        for i, (key, value) in enumerate(native_value.items()):
            buffer.append(f"{LF}{inner_indent_prefix}")
            buffer.append(f'"{key}": ')
            if isinstance(value, YuiValue):
                buffer.append(value.stringfy(inner_indent_prefix, width=width))
            else:
                buffer.append(f'{YuiType.arrayview_s(value)}')
            if i < len(native_value) - 1:
                buffer.append(",")
        buffer.append(f"{LF}{indent_prefix}}}")
        return ''.join(buffer)

    def equals(self, left_node: Any, right_node: Any) -> bool:
        left_native = YuiType.to_native(left_node)
        right_native = YuiType.to_native(right_node)
        if not isinstance(right_native, dict):
            return False
        if set(left_native.keys()) != set(right_native.keys()):
            return False
        return all(_array_equal(left_native[k], right_native[k]) for k in left_native)



YuiType.NullType = YuiNullType()
YuiType.BooleanType = YuiBooleanType()
YuiType.IntType = YuiIntType()
YuiType.FloatType = YuiFloatType()
YuiType.NumberType = YuiNumberType()
YuiType.StringType = YuiStringType()
YuiType.ObjectType = YuiObjectType()
YuiType.ArrayType = YuiArrayType()

# bool は Python で int のサブクラスなので BooleanType を IntType より先に置く
TYPES = [YuiType.NullType, YuiType.BooleanType, YuiType.IntType, YuiType.FloatType, YuiType.NumberType, YuiType.StringType, YuiType.ArrayType, YuiType.ObjectType]

def _typing(value: Any) -> YuiType:
    """値に対応するYuiTypeを返す"""
    for ty in TYPES:
        if ty.match(value):
            return ty
    raise ValueError(f"unknown type for value: {value}")

class YuiValue(object):
    """Yui言語のデータ型"""
    native_value: Optional[Union[str, float, dict, int]]
    type: YuiType
    elements: Optional[List[Any]]

    def __init__(self, native_value: Any, type: Optional[YuiType] = None):
        """YuiValueを初期化する"""
        self.native_value = YuiType.to_native(native_value)
        self.elements = None
        self.type = _typing(native_value) if type is None else type
    
    @property
    def native(self) -> Any:
        if self.native_value is None:
            self.native_value = self.type.to_native(self.elements)
        return self.native_value

    @property
    def arrayview(self):
        if self.elements is None:
            self.elements = self.type.to_arrayview(self.native_value)
        return self.elements

    def get_item(self, node_or_index: int) -> Any:
        if YuiType.is_string(node_or_index):
            key = YuiType.matched_native(node_or_index)
            if YuiType.is_object(self):
                obj = YuiType.matched_native(self)
                return YuiType.from_arrayview(obj.get(key, YuiValue.NullValue))
        
        YuiType.IntType.match_or_raise(node_or_index)
        index = YuiType.matched_native(node_or_index)
        elements = self.arrayview
        if index < 0 or index >= len(elements):
            raise YuiError(("error", "index", f"✅<{(len(elements))}", f"❌{index}", f"🔍{elements}"), node_or_index)
        return YuiType.from_arrayview(elements[index])
    
    def set_item(self, node_or_index: int, node_or_value: Any) -> Any:
        value = YuiType.into_arrayview(node_or_value)
        if YuiType.is_string(node_or_index):
            key = YuiType.matched_native(node_or_index)
            if YuiType.is_object(self):
                obj = YuiType.matched_native(self)
                obj[key] = value
                self.elements = None
                return
        YuiType.IntType.match_or_raise(node_or_index)
        index = YuiType.matched_native(node_or_index)
        elements = self.arrayview
        if index < 0 or index >= len(elements):
            raise YuiError(("error", "index", f"✅<{(len(elements))}", f"❌{index}", f"🔍{elements}"), node_or_index)
        self.type.check_element(node_or_value)
        elements[index] = value
        self.native_value = None

    def append(self, node_or_value: Any) -> Any:
        self.type.check_element(node_or_value)
        value = YuiType.into_arrayview(node_or_value)
        self.arrayview.append(value)
        self.native_value = None

    def is_primitive(self) -> bool:
        return isinstance(self.type, (YuiNullType, YuiBooleanType, YuiIntType, YuiFloatType, YuiStringType))

    @staticmethod
    def stringfy_value(value: Any, indent_prefix: str = "", width=80) -> str:
        """YuiValue または任意の値を文字列に変換する（クラスメソッドとして呼び出し可能）"""
        if isinstance(value, YuiValue):
            return value.stringfy(indent_prefix=indent_prefix, width=width)
        return str(value)

    @property
    def array(self):
        """arrayview の別名（後方互換）"""
        return self.arrayview

    def stringfy(self, indent_prefix: str = "", arrayview: bool = False, width=80) -> str:
        if arrayview:
            elements =self.arrayview
            return YuiType.ArrayType.stringfy(elements, indent_prefix=indent_prefix, width=width)
        else:
            return self.type.stringfy(self.native, indent_prefix=indent_prefix, width=width)

    def equals(self, other_node: Any) -> bool:
        return self.type.equals(self, other_node)

    def less_than(self, other_node: Any, op = "<") -> bool:
        return self.type.less_than(self, other_node, op=op)

    def __str__(self):
        """文字列表現を返す"""
        return self.stringfy(indent_prefix=None)

    def __repr__(self):
        """デバッグ用文字列表現を返す"""
        return str(self.native)

YuiValue.NullValue = YuiValue(None, type=YuiType.NullType)
YuiValue.TrueValue = YuiValue(True, type=YuiType.BooleanType)
YuiValue.FalseValue = YuiValue(False, type=YuiType.BooleanType)

## オペレーター

@dataclass
class Operator(ABC):
    symbol: str
    comparative: bool

    def __init__(self, symbol: str, comparative: bool):
        self.symbol = symbol
        self.comparative = comparative

    def __str__(self):
        return self.symbol

    @abstractmethod
    def evaluate(self, left_node: Any, right_node: Any) -> Any:
        pass

@dataclass
class Equals(Operator):
    def __init__(self, symbol: str = "=="):
        super().__init__(symbol, comparative=False)

    def evaluate(self, left_node: Any, right_node: Any) -> bool:
        left_value = left_node
        return left_value.type.equals(left_node, right_node)

@dataclass
class NotEquals(Operator):
    def __init__(self, symbol: str = "!="):
        super().__init__(symbol, comparative=False)

    def evaluate(self, left_node: Any, right_node: Any) -> bool:
        left_value = left_node
        return not left_value.type.equals(left_node, right_node)

@dataclass
class LessThan(Operator):
    def __init__(self, symbol: str = "<"):
        super().__init__(symbol, comparative=True)

    def evaluate(self, left_node: Any, right_node: Any) -> bool:
        left_value = left_node
        return not left_value.type.equals(left_node, right_node) and \
            left_value.type.less_than(left_node, right_node, op=self.symbol)

@dataclass
class GreaterThan(Operator):
    def __init__(self, symbol: str = ">"):
        super().__init__(symbol, comparative=True)

    def evaluate(self, left_node: Any, right_node: Any) -> bool:
        left_value = left_node
        return not left_value.type.equals(left_node, right_node) and \
            not left_value.type.less_than(left_node, right_node, op=self.symbol)

@dataclass
class LessThanEquals(Operator):
    def __init__(self, symbol: str = "<="):
        super().__init__(symbol, comparative=True)

    def evaluate(self, left_node: Any, right_node: Any) -> bool:
        left_value = left_node
        return left_value.type.equals(left_node, right_node) or \
            left_value.type.less_than(left_node, right_node, op=self.symbol)

@dataclass
class GreaterThanEquals(Operator):
    def __init__(self, symbol: str = ">="):
        super().__init__(symbol, comparative=True)

    def evaluate(self, left_node: Any, right_node: Any) -> bool:
        left_value = left_node
        return left_value.type.equals(left_node, right_node) or \
            not left_value.type.less_than(left_node, right_node, op=self.symbol)
    
@dataclass
class In(Operator):
    def __init__(self, symbol: str = "in"):
        super().__init__(symbol, comparative=False)

    def evaluate(self, left_node: Any, right_node: Any) -> bool:
        left_value = left_node
        right_array = right_node.arrayview
        for element in right_array:
            if left_value.type.equals(left_node, element):
                return True
        return False

@dataclass
class NotIn(Operator):
    def __init__(self, symbol: str = "notin"):
        super().__init__(symbol, comparative=False)

    def evaluate(self, left_node: Any, right_node: Any) -> bool:
        left_value = left_node
        right_array = right_node.arrayview
        for element in right_array:
            if left_value.type.equals(left_node, element):
                return False
        return True

OPERATORS = {
    '==': Equals(),
    '!=': NotEquals(),
    '<': LessThan(),
    '>': GreaterThan(),
    '<=': LessThanEquals(),
    '>=': GreaterThanEquals(),
    'in': In(),
    'notin': NotIn(),
}

set_operators(OPERATORS)




#!/usr/bin/env python
"""
Yui言語のサンプルコード生成ツール

ASTノードを使ってサンプルコードを構築し、
CodeVisitorで異なる構文（Yui、Python風など）に変換して出力します。
"""

from typing import List, Dict
from .yuiast import (
    ConstNode, NumberNode, StringNode, NameNode, ArrayNode, ObjectNode, MinusNode, ArrayLenNode,
    AssignmentNode, IncrementNode, DecrementNode, AppendNode,
    BlockNode, IfNode, RepeatNode, BreakNode, PassNode,
    FuncDefNode, FuncAppNode, ReturnNode,
    PrintExpressionNode,
    BinaryNode, GetIndexNode, AssertNode
)
from .yuicoding import CodingVisitor
from .yuisyntax import load_syntax


class YuiExample:
    """Yui言語のサンプルコード生成クラス"""

    def __init__(self, name: str, description: str, ast_node):
        self.name = name
        self.description = description
        self.ast_node = ast_node

    def generate(self, syntax: str = 'yui') -> str:
        visitor = CodingVisitor(syntax)
        return visitor.emit(self.ast_node)

def example_hello_world():
    statements = [
        PassNode(comment='Print "Hello, world!"'),
        PrintExpressionNode(StringNode("Hello, world!")),
    ]
    return YuiExample(
        name="hello_world",
        description="Print 'Hello, world!'",
        ast_node=BlockNode(statements, top_level=True)
    )

def example_variables():
    """基本的な算術演算のサンプル"""
    statements = [
        PassNode(comment="Define variables x and y"),
        AssignmentNode(NameNode("x"),NumberNode(1)),
        AssignmentNode(NameNode("y"),MinusNode(NumberNode(2))),
        PassNode(comment="Increment x"),
        IncrementNode(NameNode("x")),
        PassNode(comment="Decrement y"),
        DecrementNode(NameNode("y")),
        PassNode(comment="Test that x is 2 and y is -3"),
        AssertNode(NameNode("x"), NumberNode(2)),
        AssertNode(NameNode("y"), MinusNode(NumberNode(3))),
    ]
    return YuiExample(
        name="variables",
        description="Basic variable definition and increment/decrement",
        ast_node=BlockNode(statements, top_level=True)
    )

def example_loop():
    """ループのサンプル"""
    statements = [
        PassNode(comment="Loop 10 times and break at 5"),
        AssignmentNode(NameNode("count"),NumberNode(0)),
        RepeatNode(
            NumberNode(10),
            BlockNode([
                IncrementNode(NameNode("count")),
                IfNode(NameNode("count"), "==", NumberNode(5), BlockNode(BreakNode())),
            ]),
        ),
        PassNode(comment="Test that count is 5"),
        AssertNode(NameNode("count"), NumberNode(5)),
    ]
    return YuiExample(
        name="loop",
        description="Loop 10 times and break at 5",
        ast_node=BlockNode(statements, top_level=True)
    )

def example_nested_conditional_branches():
    """ネストした条件分岐のサンプル"""
    then_block = IncrementNode(NameNode("y"))
    else_block = IncrementNode(NameNode("z"))
    statements = [
        PassNode(comment="Test nested conditions on x and y"),
        AssignmentNode(NameNode("x"), NumberNode(1)),
        AssignmentNode(NameNode("y"), NumberNode(2)),
        AssignmentNode(NameNode("z"), NumberNode(3)),
        PassNode(comment="If x is 0, check y and increment y or z accordingly"),
        IfNode(NameNode("x"),"==", NumberNode(0),
            BlockNode(IfNode(NameNode("y"),"==", NumberNode(1),then_block, else_block)),
            BlockNode(IfNode(NameNode("y"),"==", NumberNode(2),then_block, else_block))
        ),
        PassNode(comment="Test that y was incremented and z was not"),
        AssertNode(NameNode("y"), NumberNode(3)),
    ]
    return YuiExample(
        name="nested_conditional_branches",
        description="Nested conditional branching",
        ast_node=BlockNode(statements, top_level=True)
    )

def example_comparisons():
    then_block = IncrementNode(NameNode("y"))
    else_block = IncrementNode(NameNode("z"))

    statements = [
        PassNode(comment="Various comparisons on x"),
        AssignmentNode(NameNode("x"), NumberNode(1)),
        AssignmentNode(NameNode("y"), NumberNode(0)),
        AssignmentNode(NameNode("z"), NumberNode(0)),
        PassNode(comment="Is x equal to 1?"),
        IfNode(NameNode("x"),"==", NumberNode(1),then_block, else_block),
        PassNode(comment="Is x not equal to 1?"),
        IfNode(NameNode("x"),"!=", NumberNode(1),then_block, else_block),
        PassNode(comment="Is x less than 1?"),
        IfNode(NameNode("x"),"<", NumberNode(1),then_block, else_block),
        PassNode(comment="Is x greater than 1?"),
        IfNode(NameNode("x"),">", NumberNode(1),then_block, else_block),
        PassNode(comment="Is x less than or equal to 1?"),
        IfNode(NameNode("x"),"<=", NumberNode(1),then_block, else_block),
        PassNode(comment="Is x greater than or equal to 1?"),
        IfNode(NameNode("x"),">=", NumberNode(1),then_block, else_block),
        PassNode(comment="Test that all conditions were evaluated correctly"),
        AssertNode(NameNode("y"), NumberNode(3)),
        AssertNode(NameNode("z"), NumberNode(3)),
    ]
    return YuiExample(
        name="comparisons",
        description="Comparison operations",
        ast_node=BlockNode(statements, top_level=True)
    )

def example_array():
    """配列操作のサンプル"""
    statements = [
        PassNode(comment="Create an array A with elements 1, 2, 3"),
        AssignmentNode(NameNode("A"),ArrayNode([NumberNode(1),NumberNode(2),NumberNode(3)])),
        PassNode(comment="Append 0 to the end of A"),
        AppendNode(NameNode("A"),NumberNode(0)),
        PassNode(comment="Increment the first element of A"),
        IncrementNode(GetIndexNode(NameNode("A"), NumberNode(0))),
        PassNode(comment="If 2 is in A, set the first element to the fourth element"),
        IfNode(NumberNode(2), "in", NameNode("A"),
            AssignmentNode(GetIndexNode(NameNode("A"), NumberNode(0)),GetIndexNode(NameNode("A"), NumberNode(3)))
        ),
        PassNode(comment="Test that the array has 4 elements"),
        AssertNode(ArrayLenNode(NameNode("A")), NumberNode(4)),
    ]
    return YuiExample(
        name="array",
        description="Array creation and element manipulation",
        ast_node=BlockNode(statements, top_level=True)
    )

def example_strings():
    """文字列操作のサンプル"""
    statements = [
        PassNode(comment="Create a string s with value 'hello'"),
        AssignmentNode(NameNode("s"), StringNode("hello")),
        PassNode(comment="Set the first character of s to 'H'"),
        PassNode(comment="Note: strings are just the array of character codes. So we can manipulate them like arrays."),
        AssignmentNode(GetIndexNode(NameNode("s"), NumberNode(0)), GetIndexNode(StringNode("H"), 0)),
        PassNode(comment='Append " world" to s'),
        AssignmentNode(NameNode("t"), StringNode(" world")),
        AssignmentNode(NameNode("i"), NumberNode(0)),
        RepeatNode(ArrayLenNode(NameNode("t")), BlockNode([
            AppendNode(NameNode("s"), GetIndexNode(NameNode("t"), NameNode("i"))),
            IncrementNode(NameNode("i")),
        ])),
        PassNode(comment="Test that s is now 'Hello world'"),
        AssertNode(NameNode("s"), StringNode("Hello world")),
    ]
    return YuiExample(
        name="strings",
        description="String creation and manipulation",
        ast_node=BlockNode(statements, top_level=True)
    )

def example_objects():
    """オブジェクト操作のサンプル"""
    statements = [
        PassNode(comment="Create an object O with properties x and y"),
        AssignmentNode(NameNode("O"), ObjectNode({
            "x": NumberNode(0),
            "y": NumberNode(0)
        })),
        PassNode(comment="Set the x property of O to 1"),
        AssignmentNode(GetIndexNode(NameNode("O"), StringNode("x")), NumberNode(1)),
        PassNode(comment="Set the y property of O to 2"),
        AssignmentNode(GetIndexNode(NameNode("O"), StringNode("y")), NumberNode(2)),
        PassNode(comment="Test that O has properties x=1 and y=2"),
        AssertNode(GetIndexNode(NameNode("O"), StringNode("x")), NumberNode(1)),
        AssertNode(GetIndexNode(NameNode("O"), StringNode("y")), NumberNode(2)),
    ]
    return YuiExample(
        name="objects",
        description="Object creation and property manipulation",
        ast_node=BlockNode(statements, top_level=True)
    )

def example_function():
    statements = [
        PassNode(comment="Define function that adds 1"),
        FuncDefNode(
            NameNode("succ"), [NameNode("n")],
            BlockNode([
                IncrementNode(NameNode("n")),
                ReturnNode(NameNode("n"))
            ])
        ),
        AssignmentNode(NameNode("result"),
            FuncAppNode(NameNode("succ"),[NumberNode(0)])
        ),
        AssertNode(NameNode("result"), NumberNode(1)),
    ]
    return YuiExample(
        name="function",
        description="Function definition and call (increment function)",
        ast_node=BlockNode(statements, top_level=True)
    )

def example_function_no_argument():
    statements = [
        PassNode(comment="Define function that adds 1"),
        FuncDefNode(
            NameNode("zero"), [], BlockNode(ReturnNode(NumberNode(0)))
        ),
        AssertNode(FuncAppNode(NameNode("zero"), []), NumberNode(0)),
    ]
    return YuiExample(
        name="function_no_argument",
        description="Function definition and call (zero-argument function and multi-argument function)",
        ast_node=BlockNode(statements, top_level=True)
    )

def example_function_without_return():
    statements = [
        PassNode(comment="Define function that creates a point object"),
        FuncDefNode(
            NameNode("point"), [NameNode("x"), NameNode("y")], BlockNode([
                PassNode(comment="If function does not return anything, return the local environment as an object"),
            ])
        ),
        AssignmentNode(NameNode("O"), FuncAppNode(NameNode("point"), [NumberNode(0), NumberNode(0)])),
        AssertNode(GetIndexNode(NameNode("O"), StringNode("x")), NumberNode(0)),
    ]
    return YuiExample(
        name="function_without_return",
        description="Function definition and call (function without return value)",
        ast_node=BlockNode(statements, top_level=True)
    )

def example_recursive_function():
    """関数のサンプル"""
    statements = [
        PassNode(comment="Define recursive function that computes factorial"),
        FuncDefNode(
            NameNode("fact"), [NameNode("n")],
            BlockNode([
                IfNode(NameNode("n"), "==", NumberNode(0),
                    BlockNode([ReturnNode(NumberNode(1))]),
                    BlockNode([
                        PassNode(comment="Yui does not have arithmetic operators."),
                        ReturnNode(FuncAppNode(NameNode("multiplex"), 
                        [NameNode("n"), FuncAppNode(NameNode("fact"), 
                                                    [FuncAppNode(NameNode("decrease"), [NameNode("n")])])]))
                    ])
                )
            ])
        ),
        PassNode(comment="multiplex(a, b) function for a * b."),
        FuncDefNode(
            NameNode("multiplex"), [NameNode("a"), NameNode("b")], BlockNode([
                AssignmentNode(NameNode("result"), NumberNode(0)),
                RepeatNode(NameNode("b"), BlockNode([
                    RepeatNode(NameNode("a"), BlockNode([
                        IncrementNode(NameNode("result"))
                    ]))
                ])),
                ReturnNode(NameNode("result"))
            ])
        ),
        PassNode(comment="decrease(n) function for n-1."),
        FuncDefNode(
            NameNode("decrease"), [NameNode("n")], BlockNode([
                DecrementNode(NameNode("n")),
                ReturnNode(NameNode("n"))
            ])
        ),
        PassNode(comment="Test fact(0) is 1"),
        AssertNode(FuncAppNode(NameNode("fact"),[NumberNode(0)]), NumberNode(1)),
        PassNode(comment="Test that fact(5) is 120"),
        AssertNode(FuncAppNode(NameNode("fact"),[NumberNode(5)]), NumberNode(120)),
    ]
    return YuiExample(
        name="recursive_function",
        description="Recursive function definition and call (factorial function)",
        ast_node=BlockNode(statements, top_level=True)
    )


def example_float_add():
    """float 配列の加算（stdlib なし）

    float の内部表現: [sign, d1, d2, d3, d4, d5, d6, d7]
      sign: 1 (正) または -1 (負)
      d1..d7: abs(value) * 1e6 の各桁 (合計7桁)
      例) 3.14 → [1, 3, 1, 4, 0, 0, 0, 0]
         -2.5 → [-1, 2, 5, 0, 0, 0, 0, 0]

    float_add(a, b): 同符号の float 配列を足し合わせる
    アルゴリズム: i=7 から i=1 まで逆順に桁ごとに加算し繰り上がりを伝播する
    """
    statements = [
        PassNode(comment="float format: [sign, d1..d7]  sign=1 or -1, d1..d7 = abs(x)*1e6 digits"),
        PassNode(comment="float_add(a, b): add two same-sign float arrays (no stdlib)"),
        FuncDefNode(
            NameNode("float_add"), [NameNode("a"), NameNode("b")],
            BlockNode([
                # result = [1, 0, 0, 0, 0, 0, 0, 0]  (sign は後で a[0] を設定)
                AssignmentNode(NameNode("result"), ArrayNode([
                    GetIndexNode(NameNode("a"), NumberNode(0)),
                    NumberNode(0), NumberNode(0), NumberNode(0),
                    NumberNode(0), NumberNode(0), NumberNode(0), NumberNode(0),
                ])),
                AssignmentNode(NameNode("carry"), NumberNode(0)),
                AssignmentNode(NameNode("i"), NumberNode(7)),
                # i=7 から i=1 まで 7 回くり返す
                RepeatNode(NumberNode(7), BlockNode([
                    # sum = carry
                    AssignmentNode(NameNode("sum"), NameNode("carry")),
                    # sum を a[i] 回増やす
                    RepeatNode(GetIndexNode(NameNode("a"), NameNode("i")), BlockNode([
                        IncrementNode(NameNode("sum")),
                    ])),
                    # sum を b[i] 回増やす
                    RepeatNode(GetIndexNode(NameNode("b"), NameNode("i")), BlockNode([
                        IncrementNode(NameNode("sum")),
                    ])),
                    # carry = 0
                    AssignmentNode(NameNode("carry"), NumberNode(0)),
                    # もし sum >= 10 ならば: carry=1, sum -= 10
                    IfNode(NameNode("sum"), ">=", NumberNode(10), BlockNode([
                        IncrementNode(NameNode("carry")),
                        RepeatNode(NumberNode(10), BlockNode([
                            DecrementNode(NameNode("sum")),
                        ])),
                    ])),
                    # result[i] = sum
                    AssignmentNode(
                        GetIndexNode(NameNode("result"), NameNode("i")),
                        NameNode("sum"),
                    ),
                    # i -= 1
                    DecrementNode(NameNode("i")),
                ])),
                ReturnNode(NameNode("result")),
            ])
        ),
        PassNode(comment="3.14 + 2.50 = 5.64"),
        AssignmentNode(NameNode("a"), ArrayNode([
            NumberNode(1), NumberNode(3), NumberNode(1), NumberNode(4),
            NumberNode(0), NumberNode(0), NumberNode(0), NumberNode(0),
        ])),
        AssignmentNode(NameNode("b"), ArrayNode([
            NumberNode(1), NumberNode(2), NumberNode(5), NumberNode(0),
            NumberNode(0), NumberNode(0), NumberNode(0), NumberNode(0),
        ])),
        AssignmentNode(NameNode("c"),
            FuncAppNode(NameNode("float_add"), [NameNode("a"), NameNode("b")])),
        PassNode(comment="c == [1, 5, 6, 4, 0, 0, 0, 0]  (5.640000)"),
        AssertNode(GetIndexNode(NameNode("c"), NumberNode(0)), NumberNode(1)),
        AssertNode(GetIndexNode(NameNode("c"), NumberNode(1)), NumberNode(5)),
        AssertNode(GetIndexNode(NameNode("c"), NumberNode(2)), NumberNode(6)),
        AssertNode(GetIndexNode(NameNode("c"), NumberNode(3)), NumberNode(4)),
        AssertNode(GetIndexNode(NameNode("c"), NumberNode(4)), NumberNode(0)),
        PassNode(comment="1.99 + 1.01 = 3.00  (tests carry propagation)"),
        AssignmentNode(NameNode("a"), ArrayNode([
            NumberNode(1), NumberNode(1), NumberNode(9), NumberNode(9),
            NumberNode(0), NumberNode(0), NumberNode(0), NumberNode(0),
        ])),
        AssignmentNode(NameNode("b"), ArrayNode([
            NumberNode(1), NumberNode(1), NumberNode(0), NumberNode(1),
            NumberNode(0), NumberNode(0), NumberNode(0), NumberNode(0),
        ])),
        AssignmentNode(NameNode("c"),
            FuncAppNode(NameNode("float_add"), [NameNode("a"), NameNode("b")])),
        PassNode(comment="c == [1, 3, 0, 0, 0, 0, 0, 0]  (3.000000)"),
        AssertNode(GetIndexNode(NameNode("c"), NumberNode(1)), NumberNode(3)),
        AssertNode(GetIndexNode(NameNode("c"), NumberNode(2)), NumberNode(0)),
        AssertNode(GetIndexNode(NameNode("c"), NumberNode(3)), NumberNode(0)),
    ]
    return YuiExample(
        name="float_add",
        description="Add two same-sign floats as digit arrays (no stdlib)",
        ast_node=BlockNode(statements, top_level=True)
    )


def example_null_assignment():
    """null の代入と比較"""
    statements = [
        PassNode(comment="Assign null to a variable"),
        AssignmentNode(NameNode("x"), ConstNode(None)),
        PassNode(comment="Test that x is null"),
        AssertNode(NameNode("x"), ConstNode(None)),
    ]
    return YuiExample(
        name="null_assignment",
        description="Assign null to a variable and compare",
        ast_node=BlockNode(statements, top_level=True)
    )


def example_boolean_assignment():
    """true / false の代入とアサート"""
    statements = [
        PassNode(comment="Assign true and false to variables"),
        AssignmentNode(NameNode("t"), ConstNode(True)),
        AssignmentNode(NameNode("f"), ConstNode(False)),
        PassNode(comment="Test that t is true and f is false"),
        AssertNode(NameNode("t"), ConstNode(True)),
        AssertNode(NameNode("f"), ConstNode(False)),
    ]
    return YuiExample(
        name="boolean_assignment",
        description="Assign true/false to variables and compare",
        ast_node=BlockNode(statements, top_level=True)
    )


def example_boolean_branch():
    """boolean で条件分岐"""
    statements = [
        PassNode(comment="Branch on a boolean value"),
        AssignmentNode(NameNode("flag"), ConstNode(True)),
        AssignmentNode(NameNode("result"), NumberNode(0)),
        IfNode(NameNode("flag"), "==", ConstNode(True),
            BlockNode(AssignmentNode(NameNode("result"), NumberNode(1))),
            BlockNode(AssignmentNode(NameNode("result"), NumberNode(2))),
        ),
        PassNode(comment="Test that result is 1 because flag was true"),
        AssertNode(NameNode("result"), NumberNode(1)),
    ]
    return YuiExample(
        name="boolean_branch",
        description="Conditional branch based on a boolean value",
        ast_node=BlockNode(statements, top_level=True)
    )


def example_null_check():
    """null チェック関数"""
    statements = [
        PassNode(comment="Define is_null function"),
        FuncDefNode(
            NameNode("is_null"), [NameNode("v")],
            BlockNode([
                IfNode(NameNode("v"), "==", ConstNode(None),
                    BlockNode(ReturnNode(ConstNode(True))),
                    BlockNode(ReturnNode(ConstNode(False))),
                )
            ])
        ),
        PassNode(comment="Test is_null with null and non-null values"),
        AssertNode(FuncAppNode(NameNode("is_null"), [ConstNode(None)]),  ConstNode(True)),
        AssertNode(FuncAppNode(NameNode("is_null"), [NumberNode(0)]),    ConstNode(False)),
        AssertNode(FuncAppNode(NameNode("is_null"), [StringNode("")]),   ConstNode(False)),
    ]
    return YuiExample(
        name="null_check",
        description="Function that checks if a value is null",
        ast_node=BlockNode(statements, top_level=True)
    )


# すべてのサンプルを取得
def get_all_examples() -> List[YuiExample]:
    """すべてのサンプルを返す"""
    return [
        example_hello_world(),
        example_variables(),
        example_loop(),
        example_nested_conditional_branches(),
        example_comparisons(),
        example_array(),
        example_strings(),
        example_function(),
        example_function_no_argument(),
        example_function_without_return(),
        example_recursive_function(),
        example_float_add(),
        example_null_assignment(),
        example_boolean_assignment(),
        example_boolean_branch(),
        example_null_check(),
    ]


# def print_example(example: YuiExample, syntaxes: List[str] = None):
#     """サンプルを複数の構文で出力"""
#     if syntaxes is None:
#         syntaxes = ['yui']

#     print(f"\n{'='*60}")
#     print(f"サンプル名: {example.name}")
#     print(f"説明: {example.description}")
#     print(f"{'='*60}")

#     for syntax in syntaxes:
#         try:
#             import os
#             syntax_basename = os.path.basename(syntax)
#             syntax_name = syntax_basename.replace('syntax-', '').replace('.json', '')
#             code = example.generate(syntax)
#             print(f"\n--- {syntax_name} 構文 ---")
#             print(code)
#         except Exception as e:
#             print(f"\n--- {syntax_name} 構文 (エラー) ---")
#             print(f"エラー: {e}")


# def main():
#     """メイン関数 - すべてのサンプルを出力"""
#     import argparse

#     parser = argparse.ArgumentParser(
#         description='Yui言語のサンプルコード生成ツール'
#     )
#     parser.add_argument(
#         '--syntax',
#         nargs='+',
#         default=['yui', 'pylike'],
#         help='使用する構文ファイル（複数指定可）'
#     )
#     parser.add_argument(
#         '--example',
#         type=str,
#         help='特定のサンプルのみ生成（サンプル名を指定）'
#     )
#     parser.add_argument(
#         '--list',
#         action='store_true',
#         help='利用可能なサンプルの一覧を表示'
#     )

#     args = parser.parse_args()

#     examples = get_all_examples()

#     # サンプル一覧を表示
#     if args.list:
#         print("\n利用可能なサンプル:")
#         for ex in examples:
#             print(f"  {ex.name:20s} - {ex.description}")
#         return

#     # 特定のサンプルのみ生成
#     if args.example:
#         example = next((ex for ex in examples if ex.name == args.example), None)
#         if example:
#             print_example(example, args.syntax)
#         else:
#             print(f"エラー: サンプル '{args.example}' が見つかりません")
#             print("\n利用可能なサンプル:")
#             for ex in examples:
#                 print(f"  - {ex.name}")
#         return

#     # すべてのサンプルを生成
#     for example in examples:
#         print_example(example, args.syntax)


# if __name__ == '__main__':
#     main()



import time
from typing import List, Dict, Any, Union
from types import FunctionType
from abc import ABC, abstractmethod

from .yuiast import (
    ASTNode,
    ConstNode, NumberNode, StringNode, ArrayNode, ObjectNode,
    NameNode, GetIndexNode, ArrayLenNode, MinusNode, BinaryNode, FuncAppNode,
    AssignmentNode, IncrementNode, DecrementNode, AppendNode,
    BlockNode, PassNode, PrintExpressionNode,
    IfNode, BreakNode, RepeatNode, FuncDefNode, ReturnNode,
    AssertNode, ImportNode,
)
from .yuitypes import YuiValue, YuiType, YuiError
from .yuistdlib import standard_lib
from .yuiparser import YuiParser


class YuiRuntime(object):
    """Yui言語のランタイムシステム(Visitor版)
    プログラムの実行を制御し、以下の機能を提供します：
    - プログラムのパースと実行
    - タイムアウト制御
    - 実行統計の収集（インクリメント、デクリメント、比較の回数）
    - 再帰呼び出しの追跡
    """

    environments: List[dict]
    filesystems: Dict[str, str]  # 仮想ファイルシステム
    call_frames: List[tuple]  # (func_name, args, pos, end_pos)
    increment_count: int
    decrement_count: int
    compare_count: int
    test_passed: List[str]
    test_failed: List[tuple]
    source: str

    def __init__(self):
        """YuiRuntimeを初期化する"""
        self.environments = [{}]
        self.call_frames = []
        self.filesystems = {}
        
        self.shouldStop = False
        self.timeout = 0
        self.interactive_mode = False
        self.source = ""
        self.reset_stats()

    def reset_stats(self):
        """実行統計をリセットする"""
        self.increment_count = 0
        self.decrement_count = 0
        self.compare_count = 0
        self.test_passed = []
        self.test_failed = []

    def hasenv(self, name) -> bool:
        """現在の環境に変数が存在するか確認する"""
        for env in reversed(self.environments):
            if name in env:
                return True
        return False

    def getenv(self, name) -> Any:
        """現在の環境から変数を取得する"""
        for env in reversed(self.environments):
            if name in env:
                return env[name]
        return None

    def setenv(self, name, value) -> Any:
        """現在の環境に変数を設定する"""
        self.environments[-1][name] = value  

    def pushenv(self):
        """現在の環境に変数を設定する"""
        self.environments.append({})

    def popenv(self):
        """現在の環境に変数を設定する"""
        return self.environments.pop()
    
    def stringfy_env(self, stack=-1, indent_prefix: str = "") -> str:
        """環境をJSON形式の文字列として出力する"""
        if indent_prefix is None:
            indent_prefix = ""
            inner_indent_prefix = None
            LF = ""
        else:
            inner_indent_prefix = indent_prefix + "  "
            LF = "\n"
        lines = [f"{indent_prefix}<{self.stringfy_call_frames(stack=stack)}>{LF}{{"]
        for i, (key, value) in enumerate(self.environments[stack].items()):
            if key.startswith("@"): continue 
            lines.append(f"{LF}{indent_prefix}  \"{key}\": ")
            lines.append(f"{YuiValue.stringfy_value(value, inner_indent_prefix)}")
            if i < len(self.environments[stack]) - 1:
                lines.append(", ")
        lines.append(f"{LF}{indent_prefix}}}")
        return ''.join(lines)

    def push_call_frame(self, func_name: str, args: List[Any], node):
        """関数呼び出しフレームをスタックに追加"""
        self.call_frames.append((func_name, args, node))

    def pop_call_frame(self):
        """関数呼び出しフレームをスタックから削除"""
        self.call_frames.pop()

    def stringfy_call_frames(self, stack=-1)->str:
        """スタックトレースを文字列として出力する"""
        if len(self.call_frames) == 0:
            return "global"
        call_frame = self.call_frames[stack]
        args = ", ".join(YuiValue.stringfy_value(arg, indent_prefix=None) for arg in call_frame[1])
        return f"{call_frame[0]}({args})]"

    def check_recursion_depth(self):
        """再帰呼び出しの深さをチェック"""
        if len(self.call_frames) > 512:
            args = ", ".join(str(arg) for arg in self.call_frames[-1][1])
            snippet = f"{self.call_frames[-1][0]}({args})"
            raise YuiError(("error", "recursion", f"🔍{snippet}"), self.call_frames[-1][2])

    def update_variable(self, name: str, env: Dict[str, Any], pos: int):
        """変数更新時のフック（サブクラスでオーバーライド可能）"""
        pass

    def count_inc(self):
        """インクリメント操作のカウントを増やす"""
        self.increment_count += 1

    def count_dec(self):
        """デクリメント操作のカウントを増やす"""
        self.decrement_count += 1

    def count_compare(self):
        """比較操作のカウントを増やす"""
        self.compare_count += 1

    def load(self, function: FunctionType):
        """Python関数をYui関数として読み込む"""
        return NativeFunction(function)

    def print(self, value: Any, node: ASTNode):
        """値を出力する"""
        line, _, snippet = node.extract()
        if self.interactive_mode or isinstance(node, StringNode):
            print(f"{value}")
        elif isinstance(node, FuncAppNode):
            print(f"#line: {line} {snippet.strip()}\n>>> {node.snippet}\n{value}")
        else:
            print(f"#line: {line}\n>>> {snippet.strip()}\n{value}")

    def start(self, timeout: int = 30):
        """実行を開始する"""
        self.shouldStop = False
        self.timeout = timeout
        self.startTime = time.time()

    def check_execution(self, node):
        """実行状態をチェックする"""
        # 手動停止フラグのチェック
        if self.shouldStop:
            raise YuiError(('interruptted'), node)

        # タイムアウトチェック
        if self.timeout > 0 and (time.time() - self.startTime) > self.timeout:
            raise YuiError(("error", "timeout", f"❌{self.timeout}[sec]", f"✅{self.timeout}[sec]"), node)

    def exec(self, source: str, syntax: Union[str,dict] = 'yui', timeout: int = 30, eval_mode: bool = True):
        """Yuiプログラムを実行する"""
        self.source = source

        # パースして実行
        parser = YuiParser(syntax)
        program = parser.parse(source)
        try:
            self.start(timeout)
            value = program.evaluate(self)
        except YuiError as e:
            e.runtime = self
            raise e

        # 結果を返す
        return YuiType.yui_to_native(value) if eval_mode else self.environments[-1]


    ## visitor

    def evaluate(self, node: ASTNode):
        """メインエントリーポイント。node.visit(self) に委譲する"""
        return node.visit(self)

    # ──────────────────────────────────────────────────────────
    # リテラル・値ノード
    # ──────────────────────────────────────────────────────────

    def visitConstNode(self, node: ConstNode):
        if node.native_value is True:
            return YuiValue.TrueValue
        if node.native_value is False:
            return YuiValue.FalseValue
        return YuiValue.NullValue

    def visitNumberNode(self, node: NumberNode):
        return YuiValue(node.native_value)

    def visitStringNode(self, node: StringNode):
        if isinstance(node.contents, str):
            return YuiValue(node.contents)
        parts = []
        for content in node.contents:
            if isinstance(content, str):
                parts.append(content)
            else:
                value = content.visit(self)
                parts.append(f'{YuiType.yui_to_native(value)}')
        return YuiValue(''.join(parts))

    def visitArrayNode(self, node: ArrayNode):
        array_value = YuiValue([])
        for element in node.elements:
            v = element.visit(self)
            array_value.append(v)
        return array_value

    def visitObjectNode(self, node: ObjectNode):
        object_value = YuiValue({})
        for i in range(0, len(node.elements), 2):
            key = node.elements[i].visit(self)
            val = node.elements[i + 1].visit(self)
            object_value.set_item(key, val)
        return object_value

    # ──────────────────────────────────────────────────────────
    # 変数参照・演算ノード
    # ──────────────────────────────────────────────────────────

    def visitNameNode(self, node: NameNode):
        if not self.hasenv(node.name):
            raise YuiError(("undefined", "variable", f"❌{node.name}"), node)
        return self.getenv(node.name)

    def visitGetIndexNode(self, node: GetIndexNode):
        collection = node.collection.visit(self)
        index = node.index_node.visit(self)
        return collection.get_item(index)

    def visitArrayLenNode(self, node: ArrayLenNode):
        value = node.element.visit(self)
        return YuiValue(len(value.arrayview))

    def visitMinusNode(self, node: MinusNode):
        value = node.element.visit(self)
        YuiType.NumberType.match_or_raise(value)
        return YuiValue(-YuiType.matched_native(value))

    def visitBinaryNode(self, node: BinaryNode):
        raise YuiError(
            ("error", "internal", f"🔍{node.operator} operator is not implemented"),
            node,
        )

    def visitFuncAppNode(self, node: FuncAppNode):
        name = f'@{node.name_node.name}'
        if not self.hasenv(name):
            raise YuiError(("undefined", "function", f"❌{node.name_node.name}"), node.name_node)
        function = self.getenv(name)
        if not isinstance(function, YuiFunction):
            raise YuiError(("error", "type", "✅<function>", f"❌{function}"), node.name_node)

        # 引数を訪問した直後に値を確定させる（再帰で同一ノードが上書きされる問題を防ぐ）
        arg_values = [arg_node.visit(self) for arg_node in node.arguments]

        if node.snippet == '':
            args = ', '.join(str(v) for v in arg_values)
            node.snippet = f'{node.name_node}({args})'

        return function.call(arg_values, node, self)

    # ──────────────────────────────────────────────────────────
    # 代入・変更ノード
    # ──────────────────────────────────────────────────────────

    def visitAssignmentNode(self, node: AssignmentNode):
        if not hasattr(node.variable, 'update'):
            raise YuiError(("expected", "variable", f"❌{node.variable}"), node.variable)
        value = node.expression.visit(self)
        node.variable.update(value, self)

    def visitIncrementNode(self, node: IncrementNode):
        value = node.variable.visit(self)
        YuiType.IntType.match_or_raise(value)
        node.variable.update(YuiValue(YuiType.matched_native(value) + 1), self)
        self.count_inc()

    def visitDecrementNode(self, node: DecrementNode):
        value = node.variable.visit(self)
        YuiType.IntType.match_or_raise(value)
        node.variable.update(YuiValue(YuiType.matched_native(value) - 1), self)
        self.count_dec()

    def visitAppendNode(self, node: AppendNode):
        array = node.variable.visit(self)
        value = node.expression.visit(self)
        array.append(value)

    # ──────────────────────────────────────────────────────────
    # 制御構造ノード
    # ──────────────────────────────────────────────────────────

    def visitBlockNode(self, node: BlockNode):
        for statement in node.statements:
            statement.visit(self)

    def visitIfNode(self, node: IfNode):
        left = node.left.visit(self)
        right = node.right.visit(self)
        result = node.operator.evaluate(left, right)
        self.count_compare()
        if result:
            node.then_block.visit(self)
        elif node.else_block:
            node.else_block.visit(self)

    def visitBreakNode(self, node: BreakNode):
        raise YuiBreakException()

    def visitPassNode(self, node: PassNode):
        pass

    def visitRepeatNode(self, node: RepeatNode):
        count_value = node.count_node.visit(self)
        YuiType.IntType.match_or_raise(count_value)
        count = YuiType.matched_native(count_value)
        try:
            for _ in range(abs(count)):
                self.check_execution(node)
                node.block_node.visit(self)
        except YuiBreakException:
            pass

    def visitReturnNode(self, node: ReturnNode):
        value = node.expression.visit(self)
        raise YuiReturnException(value)

    def visitFuncDefNode(self, node: FuncDefNode):
        """LocalFunctionV を登録することで、関数本体も visitor チェーンで評価される"""
        params = [p.name for p in node.parameters]
        function = LocalFunctionV(node.name_node.name, params, node.body)
        self.setenv(f'@{node.name_node.name}', function)
        return function

    # ──────────────────────────────────────────────────────────
    # 出力・テストノード
    # ──────────────────────────────────────────────────────────

    def visitPrintExpressionNode(self, node: PrintExpressionNode):
        if isinstance(node.expression, MinusNode):
            return node.expression.element.visit(self)
        if isinstance(node.expression, FuncAppNode):
            node.expression.snippet = ''
        value = node.expression.visit(self)
        self.print(value, node.expression)
        return value

    def visitAssertNode(self, node: AssertNode):
        tested = reference_value = None
        try:
            tested = node.test.visit(self)
            reference_value = node.reference.visit(self)
            if tested.type.equals(tested, reference_value):
                self.test_passed.append(str(node.test))
                return
        except YuiError:
            raise
        except Exception:
            pass
        raise YuiError(
            ("failed", "test", f"🔍{node.test}", f"❌{tested}", f"✅{reference_value}"),
            node,
        )

    def visitImportNode(self, node: ImportNode):
        """ライブラリを環境に追加する"""
        modules = []
        if node.module_name is None:
            standard_lib(modules)
        for names, func in modules:
            for name in names.split('|'): # 多言語関数名
                self.setenv(f'@{name}', NativeFunction(func))

    # # ──────────────────────────────────────────────────────────
    # # フォールバック
    # # ──────────────────────────────────────────────────────────

    # def visitASTNode(self, node: ASTNode):
    #     """visitXxxNode が未定義のノードは evaluate にフォールバック"""
    #     return node.evaluate(self)


class YuiFunction(ABC):
    name: str
    def __init__(self, name: str):
        self.name = name

    @abstractmethod
    def call(self, arguments: List[Any], node: ASTNode, runtime: 'YuiRuntime') -> YuiValue:
        pass

class LocalFunction(YuiFunction):
    """ユーザが関数の型"""
    
    def __init__(self, name: str, parameters: List[str], body: ASTNode):
        super().__init__(name)
        self.parameters = parameters
        self.body = body

    def call(self, arg_values: List[Any], node: ASTNode, runtime: 'YuiRuntime') -> YuiValue:
        runtime.pushenv()
        if len(self.parameters) != len(arg_values):
            raise YuiError(
                ("mismatch", "arguments", f"✅{len(self.parameters)}", f"❌{len(arg_values)}"), node)
        for param_name, value in zip(self.parameters, arg_values):
            runtime.setenv(param_name, value)
        try:
            runtime.push_call_frame(self.name, arg_values, node)
            runtime.check_recursion_depth()
            self.body.evaluate(runtime)
        except YuiReturnException as e:
            if e.value is not None:
                runtime.pop_call_frame()
                runtime.popenv()
                return e.value
        runtime.pop_call_frame()
        return YuiValue(runtime.popenv())


class LocalFunctionV(LocalFunction):
    """visitor パターン版 LocalFunction。body.visit(runtime) で実行する。"""

    def call(self, arg_values: List[Any], node: ASTNode, runtime: 'YuiRuntime') -> YuiValue:
        runtime.pushenv()
        if len(self.parameters) != len(arg_values):
            raise YuiError(
                ("mismatch", "arguments", f"✅{len(self.parameters)}", f"❌{len(arg_values)}"), node)
        for param_name, value in zip(self.parameters, arg_values):
            runtime.setenv(param_name, value)
        try:
            runtime.push_call_frame(self.name, arg_values, node)
            runtime.check_recursion_depth()
            self.body.visit(runtime)
        except YuiReturnException as e:
            if e.value is not None:
                runtime.pop_call_frame()
                runtime.popenv()
                return e.value
        runtime.pop_call_frame()
        return YuiValue(runtime.popenv())


class NativeFunction(YuiFunction):
    """ネイティブ関数の型"""
    is_ffi: bool = True
    name: str
    function: callable

    def __init__(self, function: callable, is_ffi=False):
        super().__init__(function.__name__)
        self.function = function
        self.is_ffi = is_ffi

    def call(self, arg_values: List[Any], node: ASTNode, runtime: 'YuiRuntime') -> YuiValue:
        try:
            result = self.function(*arg_values)
            return result if isinstance(result, YuiValue) else YuiValue(result)
        except YuiError as e:
            if e.error_node is None:
                e.error_node = node
            raise e
        except Exception as e:
            raise YuiError(("error", "internal", f"🔍{self.name}", f"⚠️ {e}"), node)

class YuiBreakException(RuntimeError):
    """ループを抜けるための例外"""
    def __init__(self):
        pass

class YuiReturnException(RuntimeError):
    """関数から値を返すための例外"""
    def __init__(self, value=None):
        self.value = value



#!/usr/bin/env python
"""
Yui Language CLI Interface
"""

import sys
import json
import os
import argparse
from typing import Dict, Any

try:
    import readline
    READLINE_AVAILABLE = True
except ImportError:
    READLINE_AVAILABLE = False

from yuichan import __version__
from .yuiparser import YuiParser
from .yuicoding import CodingVisitor
from .yuisyntax import load_syntax, generate_bnf, list_syntax_names, find_matching_syntaxes
from .yuiruntime import YuiRuntime
from .yuitypes import YuiError
from .yuiast import IncrementNode, NameNode
from . import yuiexample
from . import message as _message


def main(argv=None):
    """Main entry point"""
    if argv is None:
        argv = sys.argv[1:]

    parser = argparse.ArgumentParser(
        prog='yui',
        description=f'Yui Language Interpreter v{__version__}',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples:
  yui --syntax yui file.yui                          # Execute a file
  yui --syntax yui -i                                # Interactive mode
  yui --syntax yui --input env.json file.yui         # Load environment and execute
  yui --syntax yui file.yui --output result.json     # Save environment after execution
  yui --syntax yui --convert-to pylike file.yui      # Convert a file (saved to pylike/)
  yui --syntax yui --convert-to pylike *.yui         # Convert multiple files
  yui --syntax yui --convert-to pylike file.md       # Convert Markdown code blocks
  yui --syntax yui --bnf                             # Show BNF grammar for a syntax
  yui --syntax yui --pass@1 test/*.yui               # Run tests and show pass rate
  yui --syntax yui --test-examples                   # Test built-in examples
  yui --syntax yui --make-examples                   # Generate example .yui files
  yui --list-examples                                # List available examples
  yui --list-syntax                                  # List syntax files with examples
  yui --find-syntax file.yui                        # Find syntax that can parse given files
  yui --find-syntax test/*.yui                      # Find syntax matching multiple files

Syntax files (yuichan/syntax/):
  yui      - Japanese-style natural syntax (default)
  pylike   - Python-style syntax
  emoji    - Emoji-based syntax

Error message languages (--lang):
  ja       - Japanese (default)
  (others) - raw message fallback
"""
    )

    parser.add_argument('-V', '--version', action='version',
                        version=f'Yui {__version__}')
    parser.add_argument('-i', '--interactive', action='store_true',
                        help='Start in interactive mode')
    parser.add_argument('--syntax', type=str, metavar='FILE',
                        help='Specify syntax file (JSON) - required for execution')
    parser.add_argument('--convert-to', type=str, metavar='SYNTAX',
                        help='Convert files to target syntax and save in <SYNTAX>/ directory')
    parser.add_argument('--input', type=str, metavar='FILE',
                        help='Load environment variables from JSON file')
    parser.add_argument('--output', type=str, metavar='FILE',
                        help='Save environment variables to JSON file after execution')
    parser.add_argument('--make-examples', action='store_true',
                        help='Generate sample code files (.yui)')
    parser.add_argument('--test-examples', action='store_true',
                        help='Test all examples with YuiRuntime')
    parser.add_argument('--pass@1', dest='pass_at_1', action='store_true',
                        help='Execute multiple scripts and show pass rate')
    parser.add_argument('--example', type=str, metavar='NAME',
                        help='Generate specific example only (use with --make-examples)')
    parser.add_argument('--list-examples', action='store_true',
                        help='List available examples')
    parser.add_argument('--list-syntax', action='store_true',
                        help='List available syntax files with examples')
    parser.add_argument('--find-syntax', action='store_true',
                        help='Find syntax files that can parse all given files')
    parser.add_argument('--syntax-dir', type=str, metavar='DIR',
                        help='Directory to search for syntax JSON files')
    parser.add_argument('--bnf', action='store_true',
                        help='Display BNF grammar for the specified syntax (requires --syntax)')
    parser.add_argument('--lang', type=str, metavar='LANG', default='ja',
                        help='Error message language (default: ja)')
    parser.add_argument('file', nargs='*', help='Yui file(s) to execute')

    args = parser.parse_args(argv)

    _message.set_language(args.lang)

    try:
        # List examples
        if args.list_examples:
            list_examples()
            return

        # List syntax
        if args.list_syntax:
            list_syntax(args.syntax_dir)
            return

        # Find syntax
        if args.find_syntax:
            if not args.file:
                print("Error: --find-syntax requires at least one file", file=sys.stderr)
                sys.exit(1)
            find_syntax(args.file, args.syntax_dir)
            return

        # Pass@1 mode
        if args.pass_at_1:
            if not args.syntax:
                print("Error: --syntax option is required", file=sys.stderr)
                print("Example: yui --syntax yui --pass@1 file1.yui file2.yui", file=sys.stderr)
                print("\nAvailable syntax files:", file=sys.stderr)
                print("  - yui  (Yui style)", file=sys.stderr)
                print("  - pylike   (Python style)", file=sys.stderr)
                print("  - emoji.json       (Emoji style)", file=sys.stderr)
                sys.exit(1)
            if not args.file:
                print("Error: --pass@1 requires at least one file", file=sys.stderr)
                sys.exit(1)
            pass_at_1_mode(args.file, args.syntax)
            return

        # Test examples mode
        if args.test_examples:
            if not args.syntax:
                print("Error: --syntax option is required", file=sys.stderr)
                print("Example: yui --syntax yui --test-examples", file=sys.stderr)
                print("\nAvailable syntax files:", file=sys.stderr)
                print("  - yui  (Yui style)", file=sys.stderr)
                print("  - pylike   (Python style)", file=sys.stderr)
                print("  - emoji.json       (Emoji style)", file=sys.stderr)
                sys.exit(1)
            test_examples(args.syntax)
            return

        # Example generation mode
        if args.make_examples:
            if not args.syntax:
                print("Error: --syntax option is required", file=sys.stderr)
                print("Example: yui --syntax yui --make-examples", file=sys.stderr)
                print("\nAvailable syntax files:", file=sys.stderr)
                print("  - yui  (Yui style)", file=sys.stderr)
                print("  - pylike   (Python style)", file=sys.stderr)
                print("  - emoji.json       (Emoji style)", file=sys.stderr)
                sys.exit(1)
            make_examples(args.example, args.syntax)
            return

        # Required syntax file check (execution, interactive, conversion modes)
        if not args.syntax:
            print("Error: --syntax option is required", file=sys.stderr)
            print("\nUsage:", file=sys.stderr)
            print("  yui --syntax <syntax-file> [options] [file]", file=sys.stderr)
            print("\nExamples:", file=sys.stderr)
            print("  yui --syntax yui file.yui", file=sys.stderr)
            print("  yui --syntax pylike -i", file=sys.stderr)
            print("\nAvailable syntax files:", file=sys.stderr)
            print("  - yui  (Yui style)", file=sys.stderr)
            print("  - pylike   (Python style)", file=sys.stderr)
            print("  - emoji.json       (Emoji style)", file=sys.stderr)
            sys.exit(1)

        syntax = args.syntax

        # --syntax-dir が指定されていれば、syntax 名をそのディレクトリ内で解決する
        if args.syntax_dir and not os.path.isfile(syntax):
            candidate = syntax if syntax.endswith('.json') else f"{syntax}.json"
            full_path = os.path.join(args.syntax_dir, candidate)
            if os.path.isfile(full_path):
                syntax = full_path

        # BNF display
        if args.bnf:
            terminals = load_syntax(syntax)
            print(generate_bnf(terminals))
            return

        # Initialize environment
        env = {}
        if args.input:
            env = load_env_from_json(args.input)
            print(f"Environment loaded: {args.input}")

        # Syntax conversion mode
        if args.convert_to:
            if not args.file:
                print("Error: --convert-to requires at least one input file", file=sys.stderr)
                sys.exit(1)
            for filename in args.file:
                convert_and_save(filename, syntax, args.convert_to)
            return

        # Execute file(s)
        if args.file:
            # If multiple files are provided, execute them sequentially
            if isinstance(args.file, list):
                for filename in args.file:
                    env = run_file(filename, env, syntax)
            else:
                env = run_file(args.file, env, syntax)

            # Save environment
            if args.output:
                save_env_to_json(env, args.output)
                print(f"Environment saved: {args.output}")

            return

        # Interactive mode
        if args.interactive or not args.file:
            interactive_mode(env, syntax)
            return

    except YuiError as e:
        print(f"\nError occurred:", file=sys.stderr)
        print(e.formatted_message("| "), file=sys.stderr)
        sys.exit(1)
    except FileNotFoundError as e:
        print(f"Error: File not found - {e}", file=sys.stderr)
        sys.exit(1)
    except KeyboardInterrupt:
        print("\nExiting")
        sys.exit(0)
    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)
        import traceback
        traceback.print_exc()
        sys.exit(1)


def run_file(filename: str, env: Dict[str, Any], syntax: str = 'yui') -> Dict[str, Any]:
    """Execute a file"""
    with open(filename, 'r', encoding='utf-8') as f:
        code = f.read()

    runtime = YuiRuntime()
    for key, value in env.items():
        runtime.setenv(key, value)

    runtime.exec(code, syntax, eval_mode=False)

    # Return environment (from last scope)
    result_env = {}
    if runtime.environments:
        result_env = runtime.environments[-1].copy()

    return result_env


def interactive_mode(env: Dict[str, Any], syntax: str = 'yui'):
    """Interactive mode"""
    print(f"Yui v{__version__} - Interactive Mode")
    print(f"Syntax: {syntax}")
    print("Type 'quit' or 'exit' to exit\n")

    # Setup readline history file
    history_file = os.path.expanduser("~/.yui_history")
    if READLINE_AVAILABLE:
        try:
            if os.path.exists(history_file):
                readline.read_history_file(history_file)
            readline.set_history_length(1000)
        except Exception:
            pass

    runtime = YuiRuntime()
    for key, value in env.items():
        runtime.setenv(key, value)

    try:
        while True:
            try:
                code = input(">>> ")
                if code.lower() in ['quit', 'exit']:
                    break

                code = code.strip()
                if code == "":
                    # Display environment
                    if runtime.environments and len(runtime.environments[-1]) > 0:
                        print(json.dumps(runtime.environments[-1], indent=2, ensure_ascii=False))
                    continue

                # Parse and execute
                runtime.exec(code, syntax)

            except YuiError as e:
                print(e.formatted_message("| "))
            except KeyboardInterrupt:
                print("\nExiting")
                break
            except EOFError:
                print("\nExiting")
                break
    finally:
        if READLINE_AVAILABLE:
            try:
                readline.write_history_file(history_file)
            except Exception:
                pass


def convert_and_save(input_file: str, source_syntax: str, target_syntax: str):
    """Convert a file to target syntax and save in <target_syntax>_examples/ directory"""
    with open(input_file, 'r', encoding='utf-8') as f:
        content = f.read()

    # Convert first, then save
    if input_file.endswith('.md'):
        converted = convert_markdown_to_string(content, source_syntax, target_syntax)
    else:
        parser = YuiParser(source_syntax)
        ast = parser.parse(content)
        visitor = CodingVisitor(target_syntax)
        converted = visitor.emit(ast)

    # Derive short name for directory (strip path and .json if given)
    target_name = os.path.basename(target_syntax)
    if target_name.endswith('.json'):
        target_name = target_name[:-len('.json')]

    out_dir = target_name
    os.makedirs(out_dir, exist_ok=True)

    out_filename = os.path.join(out_dir, os.path.basename(input_file))
    with open(out_filename, 'w', encoding='utf-8') as f:
        f.write(converted)
    print(f"Converted: {input_file} -> {out_filename}")


def convert_markdown_to_string(content: str, source_syntax: str, target_syntax: str) -> str:
    """Convert code blocks in Markdown content and return as string"""
    lines = content.split('\n')
    out_lines = []
    in_code_block = False
    current_block = []

    for line in lines:
        stripped = line.strip()

        if stripped.startswith("```yui") and not in_code_block:
            in_code_block = True
            current_block = []
            out_lines.append(line)
            continue

        if stripped.startswith("```") and in_code_block:
            code = '\n'.join(current_block)
            if code.strip():
                try:
                    parser = YuiParser(source_syntax)
                    ast = parser.parse(code)
                    visitor = CodingVisitor(target_syntax)
                    out_lines.append(visitor.emit(ast))
                except Exception as e:
                    out_lines.append(f"# Conversion error: {e}")
                    out_lines.append(code)
            out_lines.append(line)
            in_code_block = False
            current_block = []
            continue

        if in_code_block:
            current_block.append(line)
        else:
            out_lines.append(line)

    return '\n'.join(out_lines)


def load_env_from_json(filename: str) -> Dict[str, Any]:
    """Load environment from JSON file"""
    with open(filename, 'r', encoding='utf-8') as f:
        return json.load(f)


def save_env_to_json(env: Dict[str, Any], filename: str):
    """Save environment to JSON file"""
    with open(filename, 'w', encoding='utf-8') as f:
        json.dump(env, f, indent=2, ensure_ascii=False)


def list_examples():
    """List available examples"""
    examples = yuiexample.get_all_examples()
    print("\nAvailable examples:")
    print(f"{'Name':<20} Description")
    print("-" * 60)
    for ex in examples:
        print(f"{ex.name:<20} {ex.description}")


def list_syntax(syntax_dir: str = None):
    """List available syntax files with an increment example for each"""
    example_node = IncrementNode(NameNode("x"))

    print("\nAvailable syntax files:")
    print(f"{'Name':<12}  {'File':<20}  {'x += 1 equivalent'}")
    print("-" * 60)

    from .yuisyntax import _DEFAULT_SYNTAX_DIR
    d = syntax_dir or _DEFAULT_SYNTAX_DIR
    for name in list_syntax_names(syntax_dir):
        filename = f"{name}.json"
        try:
            terminals = load_syntax(os.path.join(d, filename))
            if not terminals.get('if-begin', ''):
                continue
            visitor = CodingVisitor(terminals)
            code = visitor.emit(example_node).strip()
            if not code:
                code = "(no increment syntax defined)"
        except Exception as e:
            code = f"(error: {e})"
        print(f"{name:<12}  {filename:<20}  {code}")


def find_syntax(files: list, syntax_dir: str = None):
    """Find syntax files that can successfully parse all given files"""
    sources = {}
    for filename in files:
        try:
            with open(filename, 'r', encoding='utf-8') as f:
                sources[filename] = f.read()
        except FileNotFoundError:
            print(f"Error: File not found - {filename}", file=sys.stderr)
            sys.exit(1)

    results = find_matching_syntaxes(sources, syntax_dir)
    print(f"\nTrying {len(results)} syntax file(s) against {len(files)} file(s)...\n")

    matched = [name for name, ok, _ in results if ok]
    for name, ok, status in results:
        print(f"  {'✓' if ok else '✗'}  {name:<12}  {status}")

    print()
    if matched:
        print(f"Matching syntax: {', '.join(matched)}")
    else:
        print("No syntax matched all files.")


def make_examples(example_name: str = None, syntax: str = 'yui'):
    """
    Generate sample code files

    Args:
        example_name: Specific example name (all examples if None)
        syntax: Syntax file to use
    """
    examples = yuiexample.get_all_examples()

    # Derive syntax short name for directory (strip path and .json if given)
    syntax_name = os.path.basename(syntax)
    if syntax_name.endswith('.json'):
        syntax_name = syntax_name[:-len('.json')]

    examples_dir = f"{syntax_name}_examples"
    os.makedirs(examples_dir, exist_ok=True)

    # Generate specific example only
    if example_name:
        example = next((ex for ex in examples if ex.name == example_name), None)
        if not example:
            print(f"Error: Example '{example_name}' not found", file=sys.stderr)
            print("\nAvailable examples:")
            for ex in examples:
                print(f"  - {ex.name}")
            sys.exit(1)

        code = example.generate(syntax)
        filename = os.path.join(examples_dir, f"{example.name}.yui")
        with open(filename, 'w', encoding='utf-8') as f:
            f.write(code)
        print(f"Generated: {filename}")
        return

    # Generate all examples
    for example in examples:
        code = example.generate(syntax)
        filename = os.path.join(examples_dir, f"{example.name}.yui")
        with open(filename, 'w', encoding='utf-8') as f:
            f.write(code)
        print(f"Generated: {filename}")

    print(f"\nAll examples generated in {examples_dir}/ directory")


def test_examples(syntax: str = 'yui'):
    """
    Test all examples with YuiRuntime

    Args:
        syntax: Syntax file to use for parsing
    """
    examples = yuiexample.get_all_examples()

    print(f"\nTesting examples with syntax: {syntax}")
    print("=" * 60)

    passed = 0
    failed = 0

    for example in examples:
        try:
            # Generate code
            code = example.generate(syntax)

            runtime = YuiRuntime()
            runtime.exec(code, syntax)

            print(f"✓ {example.name:<20} PASSED")
            passed += 1

        except YuiError as e:
            print(f"✗ {example.name:<20} FAILED")
            print(e.formatted_message("    | "))
            failed += 1

        except Exception as e:
            print(f"✗ {example.name:<20} FAILED: {e}")
            failed += 1

    print("=" * 60)
    print(f"\nResults: {passed} passed, {failed} failed")

    if failed > 0:
        sys.exit(1)


def pass_at_1_mode(files: list, syntax: str = 'yui'):
    """
    Execute multiple script files and calculate pass rate

    - Files ending with _doctest.yui are ignored as standalone targets
    - For a.yui, if a_doctest.yui exists, both are concatenated and executed together

    Args:
        files: List of .yui files to execute
        syntax: Syntax file to use for parsing
    """
    # Filter .yui files, excluding _doctest.yui
    yui_files = [f for f in files if f.endswith('.yui') and not f.endswith('_doctest.yui')]

    if not yui_files:
        print("Error: No .yui files specified", file=sys.stderr)
        sys.exit(1)

    results = []

    for filename in yui_files:
        try:
            # Read main file
            with open(filename, 'r', encoding='utf-8') as f:
                code = f.read()

            # Check for corresponding _doctest.yui
            base = filename[:-len('.yui')]
            doctest_file = base + '_doctest.yui'
            if os.path.exists(doctest_file):
                with open(doctest_file, 'r', encoding='utf-8') as f:
                    code = code + '\n' + f.read()
                label = f"{filename} + {os.path.basename(doctest_file)}"
            else:
                label = filename

            runtime = YuiRuntime()
            runtime.exec(code, syntax)

            # Success
            results.append(1)
            print(f"✓ {label}")

        except FileNotFoundError:
            # File not found
            results.append(0)
            print(f"✗ {filename} (File not found)")

        except YuiError as e:
            # Yui syntax or runtime error
            results.append(0)
            print(f"✗ {label}")
            print(e.formatted_message("  | "))

        except Exception as e:
            # Other errors
            results.append(0)
            print(f"✗ {label}")
            print(f"  | Error: {e}")

    # Calculate pass rate
    total = len(results)
    passed = sum(results)
    pass_rate = passed / total if total > 0 else 0

    # Display results
    print(f"\n{'='*50}")
    print(f"Total: {total}")
    print(f"Passed: {passed}")
    print(f"Failed: {total - passed}")
    print(f"pass@1: {pass_rate:.2%} ({passed}/{total})")
    print(f"{'='*50}")

    # Exit with error code if any failed
    if total - passed > 0:
        sys.exit(1)


if __name__ == "__main__":
    main()



import pytest
from .yuiparser import Source, parse
from .yuiast import ConstNode
from .yuitypes import YuiError
from .yuisyntax import load_syntax

yui_syntax = load_syntax('yui')


class TestParseStatementNode_Yui:

    def test_Assignment(self):
        source = Source("x = 1 # コメント")
        assignment_node = parse("@Assignment", source, pc={})
        assert str(assignment_node) == "x = 1"

    def test_AssignmentAsStatement(self):
        source = Source("x=1 # コメント")
        assignment_node = parse("@Statement", source, pc={})
        assert str(assignment_node) == "x=1"

    def test_Increment(self):
        source = Source("xを増やす")
        increment_node = parse("@Increment", source, pc={})
        assert str(increment_node) == "xを増やす"

        source = Source("xを増やす # コメント")
        increment_node = parse("@Statement", source, pc={})
        assert str(increment_node) == "xを増やす"

    def test_Increment_lookahead_error(self):
        source = Source("xに3増やす # コメント")
        with pytest.raises(YuiError) as excinfo:
            increment_node = parse("@Statement", source, pc={})
        assert "expected" in str(excinfo.value)

    def test_Decrement(self):
        source = Source("xを減らす")
        decrement_node = parse("@Decrement", source, pc={})
        assert str(decrement_node) == "xを減らす"

        source = Source("yを減らす # コメント")
        decrement_node = parse("@Statement", source, pc={})
        assert str(decrement_node) == "yを減らす"

    def test_Append(self):
        source = Source("xに10を追加する")
        append_node = parse("@Append", source, pc={})
        assert str(append_node) == "xに10を追加する"

        source = Source("xに10を追加する # コメント")
        append_node = parse("@Statement", source, pc={})
        assert str(append_node) == "xに10を追加する"

    def test_Break(self):
        source = Source("くり返しを抜ける")
        break_node = parse("@Break", source, pc={})
        assert str(break_node) == "くり返しを抜ける"

    def test_Return(self):
        source = Source("1が答え")
        return_node = parse("@Return", source, pc={})
        assert str(return_node) == "1が答え"

        source = Source("1が答え # コメント")
        return_node = parse("@Return", source, pc={})
        assert str(return_node) == "1が答え"

    def test_FuncDef_with_arg(self):
        source = Source("f = 入力x に対して {\n  xが答え\n}")
        funcdef_node = parse("@FuncDef", source, pc={})
        assert str(funcdef_node) == "f = 入力x に対して {\n  xが答え\n}"

    def test_FuncDef_no_arg(self):
        source = Source("f = 入力なし に対して {\n  1が答え\n}")
        funcdef_node = parse("@FuncDef", source, pc={})
        assert str(funcdef_node) == "f = 入力なし に対して {\n  1が答え\n}"

    def test_FuncDef_multiple_args(self):
        source = Source("f = 入力x,y に対して {\n  xが答え\n}")
        funcdef_node = parse("@FuncDef", source, pc={})
        assert str(funcdef_node) == "f = 入力x,y に対して {\n  xが答え\n}"


class TestParseBlockNode_Yui:

    def test_TopLevel(self):
        source = Source("x = 1\ny=2")
        top_level_node = parse("@TopLevel", source, pc={})
        assert str(top_level_node) == "x = 1\ny=2"

    def test_statement_separator(self):
        source = Source("x = 1;y=2")
        source.update_syntax(**{
            "statement-separator": ";",
        })
        top_level_node = parse("@TopLevel", source, pc={})
        assert str(top_level_node) == "x = 1;y=2"

    def test_TopLevel_error(self):
        with pytest.raises(YuiError) as excinfo:
            source = Source("x = 1 知らないよ")
            top_level_node = parse("@TopLevel", source, pc={})
        assert "wrong" in str(excinfo.value)
        assert "statement" in str(excinfo.value)

    def test_Block(self):
        source = Source("n回 {\n  x = 1\n  y=2\n} くり返す", pos=3)
        block_node = parse("@Block", source, pc={})
        assert str(block_node) == "{\n  x = 1\n  y=2\n}"

    def test_Block_emptyline(self):
        source = Source("n回 {\n  x = 1\n #a\n\n  y=2\n} くり返す", pos=3)
        block_node = parse("@Block", source, pc={})
        assert str(block_node) == "{\n  x = 1\n #a\n\n  y=2\n}"

    def test_Repeat(self):
        source = Source("3回くり返す {\n  x = 1\n}")
        repeat_node = parse("@Repeat", source, pc={})
        assert str(repeat_node) == "3回くり返す {\n  x = 1\n}"

    def test_If(self):
        source = Source("もしxが1ならば {\n  x = 1\n}")
        if_node = parse("@If", source, pc={})
        assert str(if_node) == "もしxが1ならば {\n  x = 1\n}"

    def test_IfAsStatement(self):
        source = Source("もしxが1ならば {\n  x = 1\n}")
        if_node = parse("@Statement", source, pc={})
        assert str(if_node) == "もしxが1ならば {\n  x = 1\n}"

    def test_If_in(self):
        source = Source("もしxがAのいずれかならば{\n  x = 1\n}")
        if_node = parse("@If", source, pc={})
        assert str(if_node) == "もしxがAのいずれかならば{\n  x = 1\n}"

    def test_If_not_in(self):
        source = Source("もしxがAのいずれでもないならば{\n  x = 1\n}")
        if_node = parse("@If", source, pc={})
        assert str(if_node) == "もしxがAのいずれでもないならば{\n  x = 1\n}"

    def test_NestedIf(self):
        source = Source("もしxが1ならば {\n  x=0\n  もしxが0ならば {\n    x=1\n  }\n}\ny=1")
        if_node = parse("@If", source, pc={})
        assert str(if_node) == "もしxが1ならば {\n  x=0\n  もしxが0ならば {\n    x=1\n  }\n}"

    def test_NestedIfAsTopLevel(self):
        source = Source("もしxが1ならば {\n  x=0\n  もしxが0ならば {\n    x=1\n  }\n}\ny=1")
        if_node = parse("@TopLevel", source, pc={})
        assert str(if_node) == "もしxが1ならば {\n  x=0\n  もしxが0ならば {\n    x=1\n  }\n}\ny=1"

    def test_Repeat_with_break(self):
        source = Source("10回くり返す{\n   countを増やす\n   もし countが 5ならば{\n      くり返しを抜ける\n   }\n}")
        repeat_node = parse("@Repeat", source, pc={})
        assert "くり返しを抜ける" in str(repeat_node)


class TestParseConstNode_Yui:

    def test_null(self):
        source = Source("値なし")
        node = parse("@Boolean", source, pc={})
        assert isinstance(node, ConstNode)
        assert node.native_value is None
        assert str(node) == "値なし"

    def test_null_english(self):
        source = Source("null")
        node = parse("@Boolean", source, pc={})
        assert isinstance(node, ConstNode)
        assert node.native_value is None
        assert str(node) == "null"

    def test_true(self):
        source = Source("真")
        node = parse("@Boolean", source, pc={})
        assert isinstance(node, ConstNode)
        assert node.native_value is True
        assert str(node) == "真"

    def test_true_english(self):
        source = Source("true")
        node = parse("@Boolean", source, pc={})
        assert isinstance(node, ConstNode)
        assert node.native_value is True
        assert str(node) == "true"

    def test_false(self):
        source = Source("偽")
        node = parse("@Boolean", source, pc={})
        assert isinstance(node, ConstNode)
        assert node.native_value is False
        assert str(node) == "偽"

    def test_false_english(self):
        source = Source("false")
        node = parse("@Boolean", source, pc={})
        assert isinstance(node, ConstNode)
        assert node.native_value is False
        assert str(node) == "false"

    def test_null_as_term(self):
        source = Source("値なし")
        node = parse("@Term", source, pc={})
        assert isinstance(node, ConstNode)
        assert node.native_value is None

    def test_true_as_term(self):
        source = Source("真")
        node = parse("@Term", source, pc={})
        assert isinstance(node, ConstNode)
        assert node.native_value is True

    def test_false_as_term(self):
        source = Source("偽")
        node = parse("@Term", source, pc={})
        assert isinstance(node, ConstNode)
        assert node.native_value is False

    def test_null_in_assignment(self):
        source = Source("x = 値なし")
        node = parse("@Assignment", source, pc={})
        assert str(node) == "x = 値なし"

    def test_true_in_assignment(self):
        source = Source("x = 真")
        node = parse("@Assignment", source, pc={})
        assert str(node) == "x = 真"



import pytest
from .yuisyntax import (
    get_example_from_pattern
)

class TestGetFirstMatchString:

    def test_basic(self):
        pattern = r'abc'
        result = get_example_from_pattern(pattern)
        assert result == 'abc'

    def test_escape(self):
        pattern = r'\d+'
        result = get_example_from_pattern(pattern)
        assert result == '1'

    def test_charclass(self):
        pattern = r'[xyz]a'
        result = get_example_from_pattern(pattern)
        assert result == 'xa'

    def test_charclass2(self):
        pattern = r'a[AB]?b'
        result = get_example_from_pattern(pattern)
        assert result == 'ab'


    def test_optional(self):
        pattern = r'colou?r'
        result = get_example_from_pattern(pattern)
        assert result == 'color'  

    def test_alternation(self):
        pattern = r'foo|bar|baz'
        result = get_example_from_pattern(pattern)
        assert result == 'foo'

    def test_space(self):
        pattern = r'foo\s+'
        result = get_example_from_pattern(pattern)
        assert result == 'foo '

    def test_bracket(self):
        pattern = r'\['
        result = get_example_from_pattern(pattern)
        assert result == '['

    def test_bar(self):
        pattern = r'\|'
        result = get_example_from_pattern(pattern)
        assert result == '|'

    def test_paren(self):
        pattern = r'\('
        result = get_example_from_pattern(pattern)
        assert result == '('


    def test_emoji(self):
        pattern = r'[⬇️⬆️]'
        result = get_example_from_pattern(pattern)
        assert result == '⬇️'
    
    def test_emoji2(self):
        pattern = r'⬇️⬆️?⬆️'
        result = get_example_from_pattern(pattern)
        assert result == '⬇️⬆️'

    def test_group(self):
        pattern = r'ab([A-Z]+)(XXX|XYZ)'
        result = get_example_from_pattern(pattern)
        assert result == 'abAXXX'

    def test_group2(self):
        pattern = r'ab([A-Z]+)?XXX'
        result = get_example_from_pattern(pattern)
        assert result == 'abXXX'



import pytest

from .yuitypes import YuiError, YuiValue, YuiType
from .yuiruntime import YuiRuntime
from .yuiast import (
    ConstNode, NumberNode, StringNode,
    ArrayNode, ObjectNode,
    NameNode,
    MinusNode, ArrayLenNode, GetIndexNode,
    AssignmentNode, IncrementNode, DecrementNode, AppendNode,
    BlockNode, IfNode, RepeatNode, BreakNode,
    FuncDefNode, FuncAppNode, ReturnNode,
    PrintExpressionNode, AssertNode,
)
from .yuiexample import get_all_examples
_all_examples = get_all_examples()

class TestLiteral:
    """式の評価に関するテストクラス"""

    def init_runtime(self):
        runtime = YuiRuntime()
        runtime.setenv("x", YuiValue(1))
        runtime.setenv("A", YuiValue([1, 2, 3]))
        return runtime  

    def test_int(self):
        runtime = self.init_runtime()
        expression = NumberNode(42)
        result = expression.evaluate(runtime)
        result = YuiType.to_native(result)
        assert result == 42

    def test_float(self):
        runtime = self.init_runtime()
        expression = NumberNode(42.0)
        result = expression.evaluate(runtime)
        result = YuiType.to_native(result)
        assert result == 42.0

    def test_string(self):
        runtime = self.init_runtime()
        expression = StringNode("A")
        result = expression.evaluate(runtime)
        result = YuiType.to_native(result)
        assert result == "A"

    def test_string_interpolation(self):
        runtime = self.init_runtime()
        expression = StringNode(["A", NumberNode(1), "B"])
        result = expression.evaluate(runtime)
        result = YuiType.to_native(result)
        assert result == "A1B"

    def test_array(self):
        runtime = self.init_runtime()
        expression = ArrayNode([1, 2, 3])
        result = expression.evaluate(runtime)
        result = YuiType.to_native(result)
        assert result == [1, 2, 3]

    def test_object(self):
        runtime = self.init_runtime()
        expression = ObjectNode([StringNode("x"), NumberNode(1), StringNode("y"), NumberNode(2), StringNode("z"), NumberNode(3)])
        result = expression.evaluate(runtime)
        result = YuiType.to_native(result)
        assert result == {"x": 1, "y": 2, "z": 3}

class TestVariable:
    """式の評価に関するテストクラス"""

    def init_runtime(self):
        runtime = YuiRuntime()
        runtime.setenv("x", YuiValue(1))
        runtime.setenv("A", YuiValue([1, 2, 3]))
        return runtime  

    def test_variable(self):
        runtime = self.init_runtime()
        expression = NameNode("x")
        result = expression.evaluate(runtime)
        result = YuiType.to_native(result)
        assert result == 1

    def test_undefined_variable(self):
        runtime = self.init_runtime()
        expression = NameNode("y")
        with pytest.raises(YuiError) as excinfo:    
            result = expression.evaluate(runtime)
            result = YuiType.to_native(result)
            assert result == 1
        assert "undefined" in str(excinfo.value.args[0])
        assert "variable" in str(excinfo.value.args[0])

class TestUnaryOperator:
    def init_runtime(self):
        runtime = YuiRuntime()
        runtime.setenv("x", YuiValue(1))
        runtime.setenv("A", YuiValue([1, 2, 3]))
        return runtime
    
    def test_minus_int(self):
        runtime = self.init_runtime()
        expression = MinusNode(NumberNode(1))
        result = expression.evaluate(runtime)
        result = YuiType.to_native(result)
        assert result == -1

    def test_minus_float(self):
        runtime = self.init_runtime()
        expression = MinusNode(NumberNode(1.0))
        result = expression.evaluate(runtime)
        result = YuiType.to_native(result)
        assert result == -1.0

    def test_minus_string(self):
        runtime = self.init_runtime()
        expression = MinusNode(StringNode("A"))
        with pytest.raises(YuiError) as excinfo:    
            result = expression.evaluate(runtime)
            result = YuiType.to_native(result)
            assert result == "A"
        assert "type" in str(excinfo.value)

    def test_length_array(self):
        runtime = self.init_runtime()
        expression = ArrayLenNode(NameNode("A"))
        result = expression.evaluate(runtime)
        result = YuiType.to_native(result)
        assert result == 3

    def test_length_string(self):
        runtime = self.init_runtime()
        expression = ArrayLenNode(StringNode("abc"))
        result = expression.evaluate(runtime)
        result = YuiType.to_native(result)
        assert result == 3

    def test_length_int(self):
        runtime = self.init_runtime()
        expression = ArrayLenNode(NumberNode(1))
        result = expression.evaluate(runtime)
        result = YuiType.to_native(result)
        assert result == 32

class TestGetIndex:
    def init_runtime(self):
        runtime = YuiRuntime()
        runtime.setenv("x", YuiValue(1))
        runtime.setenv("s", YuiValue("abc"))
        runtime.setenv("A", YuiValue([1, 2, 3]))
        runtime.setenv("O", YuiValue({"x": 1, "y": 2, "z": 3}))
        return runtime

    def test_get_array(self):
        runtime = self.init_runtime()
        expression = GetIndexNode(NameNode("A"), NumberNode(1))
        result = expression.evaluate(runtime)
        result = YuiType.to_native(result)
        assert result == 2

    def test_get_string(self):
        runtime = self.init_runtime()
        expression = GetIndexNode(NameNode("s"), NumberNode(1))
        result = expression.evaluate(runtime)
        result = YuiType.to_native(result)
        assert result == 98 # 'b'のASCIIコード

    def test_get_charcode(self):
        runtime = self.init_runtime()
        expression = GetIndexNode(StringNode("b"), NumberNode(0))
        result = expression.evaluate(runtime)
        result = YuiType.to_native(result)
        assert result == 98 # 'b'のASCIIコード

    def test_object_string(self):
        runtime = self.init_runtime()
        expression = GetIndexNode(NameNode("O"), StringNode("y"))
        result = expression.evaluate(runtime)
        result = YuiType.to_native(result)
        assert result == 2

    def test_get_array_out_of_index(self):
        runtime = self.init_runtime()
        expression = GetIndexNode(NameNode("A"), NumberNode(3))
        with pytest.raises(YuiError) as excinfo:    
            result = expression.evaluate(runtime)
        assert "index" in str(excinfo.value)

class TestAssignment:

    def init_runtime(self):
        runtime = YuiRuntime()
        runtime.setenv("x", YuiValue(1))
        runtime.setenv("s", YuiValue("abc"))
        runtime.setenv("A", YuiValue([1, 2, 3]))
        runtime.setenv("O", YuiValue({"x": 1, "y": 2, "z": 3}))
        return runtime

    def test_variable_assignment(self):
        runtime = self.init_runtime()
        statement = AssignmentNode(NameNode("A"), NumberNode(3))
        statement.evaluate(runtime)
        assert YuiType.compare(runtime.getenv("A"), 3) == 0

    def test_string_assignment(self):
        runtime = self.init_runtime()
        with pytest.raises(YuiError) as excinfo:
            statement = AssignmentNode(StringNode("A"), NumberNode(3))
            statement.evaluate(runtime)
        assert "expected" in str(excinfo.value)
        assert "variable" in str(excinfo.value)

    def test_variable_increment(self):
        runtime = self.init_runtime()
        statement = IncrementNode(NameNode("x"))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("x")) == 2

    def test_variable_decrement(self):
        runtime = self.init_runtime()
        statement = DecrementNode(NameNode("x"))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("x")) == 0

    def test_array_assignment(self):
        runtime = self.init_runtime()
        statement = AssignmentNode(GetIndexNode(NameNode("A"), NumberNode(0)), NumberNode(3))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("A").get_item(0)) == 3

    def test_array_increment(self):
        runtime = self.init_runtime()
        statement = IncrementNode(GetIndexNode(NameNode("A"), NumberNode(0)))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("A").get_item(0)) == 2

    def test_array_decrement(self):
        runtime = self.init_runtime()
        statement = DecrementNode(GetIndexNode(NameNode("A"), NumberNode(1)))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("A").get_item(1)) == 1

    def test_object_assignment(self):
        runtime = self.init_runtime()
        statement = AssignmentNode(GetIndexNode(NameNode("O"), StringNode("x")), NumberNode(3))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("O").get_item("x")) == 3

class TestAppend:

    def init_runtime(self):
        runtime = YuiRuntime()
        runtime.setenv("x", YuiValue(1))
        runtime.setenv("s", YuiValue("abc"))
        runtime.setenv("A", YuiValue([1, 2, 3]))
        runtime.setenv("O", YuiValue({"x": 1, "y": 2, "z": 3}))
        return runtime

    def test_append_array(self):
        runtime = self.init_runtime()
        statement = AppendNode(NameNode("A"), NumberNode(0))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("A").get_item(3)) == 0

    def test_append_string(self):
        runtime = self.init_runtime()
        statement = AppendNode(NameNode("s"), NumberNode(ord("d")))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("s").get_item(3)) == ord("d")

    def test_append_int(self):
        runtime = self.init_runtime()
        statement = AppendNode(NameNode("x"), NumberNode(1))
        with pytest.raises(YuiError) as excinfo:
            statement.evaluate(runtime)
            assert YuiType.to_native(runtime.getenv("x")) == 2
        assert "array" in str(excinfo.value)

class TestIfCondition:

    def init_runtime(self):
        runtime = YuiRuntime()
        runtime.setenv("x", YuiValue(1))
        runtime.setenv("s", YuiValue("abc"))
        runtime.setenv("A", YuiValue([1, 2, 3]))
        runtime.setenv("O", YuiValue({"x": 1, "y": 2, "z": 3}))
        return runtime

    def make_if_statement(self, left, operator, right):
        return IfNode(left, operator, right,
            AssignmentNode(NameNode("result"), 1),
            AssignmentNode(NameNode("result"), 0)
        )

    def test_if_eq(self):
        runtime = self.init_runtime()
        statement = self.make_if_statement(NameNode("x"), "==", NumberNode(1))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("result")) == 1
        runtime = self.init_runtime()
        statement = self.make_if_statement(NameNode("x"), "==", NumberNode(0))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("result")) == 0

    def test_if_ne(self):
        runtime = self.init_runtime()
        statement = self.make_if_statement(NameNode("x"), "!=", NumberNode(1))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("result")) == 0

        runtime = self.init_runtime()
        statement = self.make_if_statement(NameNode("x"), "!=", NumberNode(0))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("result")) == 1

    def test_if_lt(self):
        runtime = self.init_runtime()
        statement = self.make_if_statement(NameNode("x"), "<", NumberNode(0))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("result")) == 0

        runtime = self.init_runtime()
        statement = self.make_if_statement(NameNode("x"), "<", NumberNode(1))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("result")) == 0

        runtime = self.init_runtime()
        statement = self.make_if_statement(NameNode("x"), "<", NumberNode(2))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("result")) == 1

    def test_if_le(self):
        runtime = self.init_runtime()
        statement = self.make_if_statement(NameNode("x"), "<=", NumberNode(0))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("result")) == 0

        runtime = self.init_runtime()
        statement = self.make_if_statement(NameNode("x"), "<=", NumberNode(1))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("result")) == 1

        runtime = self.init_runtime()
        statement = self.make_if_statement(NameNode("x"), "<=", NumberNode(2))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("result")) == 1

    def test_if_gt(self):
        runtime = self.init_runtime()
        statement = self.make_if_statement(NameNode("x"), ">", NumberNode(0))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("result")) == 1

        runtime = self.init_runtime()
        statement = self.make_if_statement(NameNode("x"), ">", NumberNode(1))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("result")) == 0

        runtime = self.init_runtime()
        statement = self.make_if_statement(NameNode("x"), ">", NumberNode(2))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("result")) == 0

    def test_if_ge(self):
        runtime = self.init_runtime()
        statement = self.make_if_statement(NameNode("x"), ">=", NumberNode(0))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("x")) == 1

        runtime = self.init_runtime()
        statement = self.make_if_statement(NameNode("x"), ">=", NumberNode(1))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("result")) == 1

        runtime = self.init_runtime()
        statement = self.make_if_statement(NameNode("x"), ">=", NumberNode(2))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("result")) == 0

    def test_if_eq_string(self):
        runtime = self.init_runtime()
        statement = self.make_if_statement(NameNode("s"), "==", StringNode("abc"))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("result")) == 1
        runtime = self.init_runtime()
        statement = self.make_if_statement(NameNode("s"), "==", StringNode("ABC"))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("result")) == 0

    def test_if_eq_float(self):
        runtime = self.init_runtime()
        runtime.setenv("f", YuiValue(3.14))
        statement = self.make_if_statement(NameNode("f"), "==", NumberNode(3.14))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("result")) == 1
        runtime = self.init_runtime()
        runtime.setenv("f", YuiValue(3.1))
        statement = self.make_if_statement(NameNode("f"), "==", NumberNode(3.145))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("result")) == 0


class TestStatement:

    def init_runtime(self):
        runtime = YuiRuntime()
        runtime.setenv("x", YuiValue(1))
        runtime.setenv("s", YuiValue("abc"))
        runtime.setenv("A", YuiValue([1, 2, 3]))
        runtime.setenv("O", YuiValue({"x": 1, "y": 2, "z": 3}))
        return runtime

    def test_block(self):
        runtime = self.init_runtime()
        statement = BlockNode([
            AssignmentNode(NameNode("x"), NumberNode(10)),
            IncrementNode(NameNode("x")),
            IncrementNode(NameNode("x")),
        ])
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("x")) == 12
    
    def test_repeat(self):
        runtime = self.init_runtime()
        statement = RepeatNode(3,IncrementNode(NameNode("x")))
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("x")) == 4

    def test_repeat_break(self):
        runtime = self.init_runtime()
        statement = RepeatNode(
            NumberNode(10),
            BlockNode([
                IncrementNode(NameNode("x")),
                IfNode(
                    left=NameNode("x"),
                    operator="==",
                    right=NumberNode(4),
                    then_block=BlockNode([BreakNode()])
                )
            ])
        )
        statement.evaluate(runtime)
        assert YuiType.to_native(runtime.getenv("x")) == 4

    def test_print_expression(self, capsys):
        runtime = self.init_runtime()
        expression = PrintExpressionNode(
            NumberNode(5)
        )
        expression.evaluate(runtime)
        captured = capsys.readouterr()
        output = captured.out.strip()
        assert "5" in output
        assert ">>>" in output

    def test_hello_world(self, capsys):
        runtime = self.init_runtime()
        expression = PrintExpressionNode(
            StringNode("hello, world")
        )
        expression.evaluate(runtime)
        captured = capsys.readouterr()
        output = captured.out.strip()
        assert "hello, world" in output
        assert ">>>" not in output

    def test_print_funcapp(self, capsys):
        runtime = self.init_runtime()
        func_def = FuncDefNode(
            NameNode("add"),[NameNode("a"), NameNode("b")],
            BlockNode([])
        )
        func_def.evaluate(runtime)
        func_app = FuncAppNode(
            name=NameNode("add"),
            arguments=[NumberNode(1), NumberNode(2)]
        )
        expression = PrintExpressionNode(
            func_app
        )
        expression.evaluate(runtime)
        captured = capsys.readouterr()
        output = captured.out.strip()
        assert "(1, 2)" in output

    def test_assert(self):
        runtime = self.init_runtime()
        expression = AssertNode(NameNode("x"), NumberNode(1))
        expression.evaluate(runtime)
        assert len(runtime.test_passed) == 1

    def test_assert_fail(self):
        runtime = self.init_runtime()
        expression = AssertNode(NameNode("x"), NumberNode(0))
        with pytest.raises(YuiError) as excinfo:
            expression.evaluate(runtime)
            assert len(runtime.test_passed) == 0
        assert "failed" in str(excinfo.value.args[0])

class TestAssert:
    """AssertNode の型別テスト"""

    def rt(self):
        return YuiRuntime()

    # ── boolean ──────────────────────────────────────────────
    def test_bool_true(self):
        rt = self.rt()
        AssertNode(ConstNode(True), ConstNode(True)).evaluate(rt)
        assert len(rt.test_passed) == 1

    def test_bool_false(self):
        rt = self.rt()
        AssertNode(ConstNode(False), ConstNode(False)).evaluate(rt)
        assert len(rt.test_passed) == 1

    def test_bool_mismatch(self):
        rt = self.rt()
        with pytest.raises(YuiError) as exc:
            AssertNode(ConstNode(True), ConstNode(False)).evaluate(rt)
        assert "failed" in str(exc.value.args[0])

    # ── int ──────────────────────────────────────────────────
    def test_int_equal(self):
        rt = self.rt()
        AssertNode(NumberNode(42), NumberNode(42)).evaluate(rt)
        assert len(rt.test_passed) == 1

    def test_int_mismatch(self):
        rt = self.rt()
        with pytest.raises(YuiError) as exc:
            AssertNode(NumberNode(1), NumberNode(2)).evaluate(rt)
        assert "failed" in str(exc.value.args[0])

    def test_int_vs_bool(self):
        """int と bool は等値にならない"""
        rt = self.rt()
        with pytest.raises(YuiError):
            AssertNode(NumberNode(1), ConstNode(True)).evaluate(rt)

    # ── float ─────────────────────────────────────────────────
    def test_float_equal(self):
        rt = self.rt()
        AssertNode(NumberNode(3.14), NumberNode(3.14)).evaluate(rt)
        assert len(rt.test_passed) == 1

    def test_float_rounding(self):
        """小数は小数点6桁で丸めて比較"""
        rt = self.rt()
        AssertNode(NumberNode(1.0000001), NumberNode(1.0000002)).evaluate(rt)
        assert len(rt.test_passed) == 1

    def test_float_mismatch(self):
        rt = self.rt()
        with pytest.raises(YuiError) as exc:
            AssertNode(NumberNode(3.14), NumberNode(3.15)).evaluate(rt)
        assert "failed" in str(exc.value.args[0])

    # ── string ────────────────────────────────────────────────
    def test_string_equal(self):
        rt = self.rt()
        AssertNode(StringNode("hello"), StringNode("hello")).evaluate(rt)
        assert len(rt.test_passed) == 1

    def test_string_mismatch(self):
        rt = self.rt()
        with pytest.raises(YuiError) as exc:
            AssertNode(StringNode("hello"), StringNode("world")).evaluate(rt)
        assert "failed" in str(exc.value.args[0])

    # ── array ─────────────────────────────────────────────────
    def test_array_equal(self):
        rt = self.rt()
        AssertNode(ArrayNode([1, 2, 3]), ArrayNode([1, 2, 3])).evaluate(rt)
        assert len(rt.test_passed) == 1

    def test_array_nested(self):
        rt = self.rt()
        AssertNode(
            ArrayNode([ArrayNode([1, 2]), ArrayNode([3, 4])]),
            ArrayNode([ArrayNode([1, 2]), ArrayNode([3, 4])]),
        ).evaluate(rt)
        assert len(rt.test_passed) == 1

    def test_array_charcode_vs_string(self):
        """文字コード配列と文字列は等値になる"""
        rt = self.rt()
        # [72, 105] == "Hi"
        AssertNode(ArrayNode([72, 105]), StringNode("Hi")).evaluate(rt)
        assert len(rt.test_passed) == 1

    def test_array_mismatch(self):
        rt = self.rt()
        with pytest.raises(YuiError) as exc:
            AssertNode(ArrayNode([1, 2, 3]), ArrayNode([1, 2, 4])).evaluate(rt)
        assert "failed" in str(exc.value.args[0])

    def test_array_length_mismatch(self):
        rt = self.rt()
        with pytest.raises(YuiError) as exc:
            AssertNode(ArrayNode([1, 2]), ArrayNode([1, 2, 3])).evaluate(rt)
        assert "failed" in str(exc.value.args[0])

    # ── object ────────────────────────────────────────────────
    def test_object_equal(self):
        rt = self.rt()
        obj = lambda: ObjectNode([StringNode("x"), NumberNode(1), StringNode("y"), NumberNode(2)])
        AssertNode(obj(), obj()).evaluate(rt)
        assert len(rt.test_passed) == 1

    def test_object_mismatch(self):
        rt = self.rt()
        with pytest.raises(YuiError) as exc:
            AssertNode(
                ObjectNode([StringNode("x"), NumberNode(1)]),
                ObjectNode([StringNode("x"), NumberNode(2)]),
            ).evaluate(rt)
        assert "failed" in str(exc.value.args[0])


class TestFunction:

    def init_runtime(self):
        runtime = YuiRuntime()
        runtime.setenv("x", YuiValue(1))
        runtime.setenv("s", YuiValue("abc"))
        runtime.setenv("A", YuiValue([1, 2, 3]))
        runtime.setenv("O", YuiValue({"x": 1, "y": 2, "z": 3}))
        return runtime

    def test_function(self):
        runtime = self.init_runtime()
        func_def = FuncDefNode(
            NameNode("add"),[NameNode("a"), NameNode("b")],
            BlockNode([ReturnNode(MinusNode(MinusNode(NameNode("a"))))])
        )
        func_def.evaluate(runtime)
        func_app = FuncAppNode(
            name=NameNode("add"),
            arguments=[NumberNode(10), NumberNode(20)]
        )
        result = func_app.evaluate(runtime)
        assert YuiType.to_native(result) == 10

    def test_function_return_none(self):
        runtime = self.init_runtime()
        func_def = FuncDefNode(
            NameNode("add"),[NameNode("a"), NameNode("b")],
            BlockNode([
                AssignmentNode(NameNode("a"), 1),
                AssignmentNode(NameNode("c"), 3)
            ])
        )
        func_def.evaluate(runtime)
        func_app = FuncAppNode(
            name=NameNode("add"),
            arguments=[NumberNode(0), NumberNode(1)]
        )
        result = func_app.evaluate(runtime)
        assert YuiType.is_object(result)
        result = YuiType.to_native(result)
        assert result["a"] == 1
        assert result["b"] == 1
        assert result["c"] == 3

    def test_function_undefined(self):
        runtime = self.init_runtime()
        func_app = FuncAppNode(
            name=NameNode("sub"),
            arguments=[NumberNode(10), NumberNode(20)]
        )
        with pytest.raises(YuiError) as excinfo:    
            result = func_app.evaluate(runtime)
        assert "undefined" in str(excinfo.value.args[0])
        assert "function" in str(excinfo.value.args[0])
    
    def test_function_argument_mismatch(self):
        runtime = self.init_runtime()
        func_def = FuncDefNode(
            NameNode("add"),[NameNode("a"), NameNode("b")],
            BlockNode([ReturnNode(MinusNode(MinusNode(NameNode("a"))))])
        )
        func_def.evaluate(runtime)
        func_app = FuncAppNode(
            name=NameNode("add"),
            arguments=[NumberNode(10)]
        )
        with pytest.raises(YuiError) as excinfo:    
            result = func_app.evaluate(runtime)
        assert "mismatch" in str(excinfo.value.args[0])
        assert "argument" in str(excinfo.value.args[0]) 

    
class TestExample:

    def get_example_ast(self, name):
        from .yuiexample import get_all_examples
        examples = get_all_examples()
        for example in examples:
            if name in example.name:
                return example.ast_node

    def test_example_hello_world(self):
        runtime = YuiRuntime()
        program = self.get_example_ast("hello")
        program.evaluate(runtime)

    @pytest.mark.parametrize("example", _all_examples, ids=lambda e: e.name)
    def test_all_examples(self, example):
        runtime = YuiRuntime()
        example.ast_node.evaluate(runtime)

